/*
 * File:     wx_panel.h
 * Purpose:  wxPanel subwindow, for panel items (widgets/controls)
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef wx_panelh
#define wx_panelh

#include "common.h"
#include "wb_panel.h"

#ifdef wx_motif
#include <Xm/Xm.h>
#endif

#define PANEL_HSPACING  10
#define PANEL_VSPACING  8
#define PANEL_LEFT_MARGIN 4
#define PANEL_TOP_MARGIN  4

#ifdef IN_CPROTO
typedef       void    *wxPanel ;
#else

class wxItem;
class wxFrame;
class wxPanel: public wxbPanel
{
#ifdef wx_motif
  // For panel item positioning.
  int cursor_x;
  int cursor_y;
  int max_width;
  int max_height;
  int max_line_height;
  wxWindow *last_created ;
#endif
 public:
#ifdef wx_motif
//  int hSpacing;
//  int vSpacing;
  Widget borderWidget;
  Widget panelWidget;
  Widget lastWidget;
  Widget firstRowWidget;
  Bool allRelative;  // If TRUE, all widgets are relatively placed so we can do
                     // some Motif-specific things to space them better
  Bool absolutePositioning;
  int currentRow;
  int currentCol;
#endif
#ifdef wx_xview
   int x_create ;
   int y_create ;  // Where creation occurs in my father
#endif

  wxPanel(void);
  wxPanel(wxFrame *frame,
          int x=-1, int y=-1, int width=-1, int height=-1, int style=0,
          char *name = "panel");
  wxPanel(wxPanel *panel,
          int x=-1, int y=-1, int width=-1, int height=-1, int style=0,
          char *name = "panel" );

  ~wxPanel(void);

  Bool Create(wxFrame *frame,
          int x=-1, int y=-1, int width=-1, int height=-1, int style=0,
          char *name = "panel");
  Bool Create(wxPanel *panel,
          int x=-1, int y=-1, int width=-1, int height=-1, int style=0,
          char *name = "panel");

  void SetSize(int x, int y, int width, int height);
  void SetClientSize(int w, int h);
  void GetPosition(int *x, int *y);
  void Centre(int direction) ;
  virtual void Enable(Bool Flag) ;
  void AddChild(wxObject *child) ;

  // Start a new line
  void NewLine(void);
  void NewLine(int pixels);
#ifdef wx_motif
  void RealNewLine(void) ;
#endif
  // Tab specified number of pixels
  void Tab(void);
  void Tab(int pixels);

  void GetCursor(int *x, int *y);

  // Set/get horizontal spacing
  void SetHorizontalSpacing(int sp);
  int GetHorizontalSpacing(void);

  // Set/get vertical spacing
  void SetVerticalSpacing(int sp);
  int GetVerticalSpacing(void);

  // Fits the panel around the items
  void Fit(void);

  // Update next cursor position
  void AdvanceCursor(wxWindow *item);
#if wx_motif
  void RealAdvanceCursor(void);
#endif
  // If x or y are not specified (i.e. < 0), supply
  // values based on left to right, top to bottom layout.
  // Internal use only.
  void GetValidPosition(int *x, int *y);

#ifdef wx_motif
  void OptimizeLayout(void);
  void AttachWidget(wxPanel *panel, Widget formWidget,
                    int x, int y, int width, int height) ;
#endif
#ifdef wx_xview
  virtual void DragAcceptFiles(Bool accept = TRUE);
#endif
};

#endif // IN_CPROTO
#endif // wx_panelh
