/*
 * File:     wb_dc.cc
 * Purpose:  Device context implementation
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Uncomment for Borland/UNIX combination
#ifdef wx_msw
#include "wx.h"
#pragma hdrstop
#else

#include "common.h"
#include "wx_frame.h"
#include "wx_dc.h"
#include "wx_dcps.h"
#include "wx_stdev.h"
#include "wx_utils.h"
#include "wx_canvs.h"
#include "wx_dialg.h"
#include "wx_main.h"
#endif
*/

#include "wx.h"                         // MSC 7/UNIX
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>


// Some global printer parameters
char *wx_printer_command = "lpr";
char *wx_preview_command = "ghostview";
char *wx_printer_flags = NULL;
char wx_printer_file[400];
Bool wx_preview = TRUE;
Bool wx_portrait = TRUE;
float wx_printer_scale_x = 1.0;
float wx_printer_scale_y = 1.0;
float wx_printer_translate_x = 0.0;
float wx_printer_translate_y = 0.0;
Bool wx_print_to_file = FALSE;
int wxPageNumber;

void wx_draw_open_spline(wxbDC *dc, wxSpline *spline);
void wx_draw_open_spline_ps(wxPostScriptDC *dc, wxSpline *s);

void wx_quadratic_spline(float a1, float b1, float a2, float b2,
                         float a3, float b3, float a4, float b4);
void wx_clear_stack(void);
int wx_spline_pop(float *x1, float *y1, float *x2, float *y2, float *x3,
        float *y3, float *x4, float *y4);
void wx_spline_push(float x1, float y1, float x2, float y2, float x3, float y3,
          float x4, float y4);
static Bool wx_spline_add_point(int x, int y);
static void wx_spline_draw_point_array(wxbDC *dc);
wxSpline *wx_make_spline(int x1, int y1, int x2, int y2, int x3, int y3);

// Default constructor
wxbDC::wxbDC(void)
{
  __type = wxTYPE_DC;
  min_x = 0; min_y = 0; max_x = 0; max_y = 0;
  title = NULL;
}

wxbDC::~wxbDC(void)
{
  if (title)
    delete[] title;
}

void wxbDC::DrawPolygon(wxList *list, float xoffset, float yoffset)
{
  int n = list->Number();
  wxPoint *points = new wxPoint[n];
  int i = 0;
  wxNode *node = list->First();
  while (node)
  {
    wxPoint *point = (wxPoint *)node->Data();
    points[i].x = point->x;
    points[i].y = point->y;
    i ++;
    node = node->Next();
  }
  DrawPolygon(n, points, xoffset, yoffset);
  delete[] points;
}

void wxbDC::DrawLines(wxList *list, float xoffset, float yoffset)
{
  int n = list->Number();
  wxPoint *points = new wxPoint[n];
  int i = 0;
  wxNode *node = list->First();
  while (node)
  {
    wxPoint *point = (wxPoint *)node->Data();
    points[i].x = point->x;
    points[i].y = point->y;
    i ++;
    node = node->Next();
  }
  DrawLines(n, points, xoffset, yoffset);
  delete []points;
}

void wxbDC::SetTextForeground(wxColour *colour)
{
  if (!colour)
    return;
  current_text_foreground = *colour;
}

void wxbDC::SetTextBackground(wxColour *colour)
{
  if (!colour)
    return;
  current_text_background = *colour;
}

void wxbDC::SetBackgroundMode(int mode)
{
  current_bk_mode = mode;
}

void wxbDC::GetSize(float *width, float *height)
{
  if (!(min_x == 1000.0 && min_y == 1000.0 && max_x == -1000.0 && max_y == -1000.0))
  {
    *width = (float)(max_x - min_x);
    *height = (float)(max_y - min_y);
  }
  else
  {
    *width = 0.0;
    *height = 0.0;
  }
}

// Make a 3-point spline
void wxbDC::DrawSpline(float x1, float y1, float x2, float y2, float x3, float y3)
{
  wxList *point_list = new wxList;

  wxPoint *point1 = new wxPoint;
  point1->x = x1; point1->y = y1;
  point_list->Append(point1);

  wxPoint *point2 = new wxPoint;
  point2->x = x2; point2->y = y2;
  point_list->Append(point2);

  wxPoint *point3 = new wxPoint;
  point3->x = x3; point3->y = y3;
  point_list->Append(point3);

  wxSpline spline(point_list);

  wx_draw_open_spline(this, &spline);
  spline.DeletePoints();
}

void wxbDC::DrawSpline(wxList *list)
{
  wxSpline spline(list);

  wx_draw_open_spline(this, &spline);
}

void wxbDC::SetLogicalOrigin(float x, float y)
{
  logical_origin_x = x;
  logical_origin_y = y;
}

void wxbDC::SetDeviceOrigin(float x, float y)
{
  device_origin_x = x;
  device_origin_y = y;
}

// For use by wxWindows only, unless custom units are required.
void wxbDC::SetLogicalScale(float x, float y)
{
  logical_scale_x = x;
  logical_scale_y = y;
}

void wxbDC::CalcBoundingBox(float x, float y)
{
  if (x < min_x)
    min_x = x;
  if (y < min_y)
    min_y = y;

  if (x > max_x)
    max_x = x;

  if (y > max_y)
    max_y = y;
}

/***************************************************************************
 * Spline code taken from xfig - see below   
 *
 * FIG : Facility for Interactive Generation of figures
 * Copyright (c) 1985 by Supoj Sutanthavibul
 *
 * "Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty."
 *
 **************************************************************************/

wxList wx_spline_point_list;

void wx_draw_open_spline(wxbDC *dc, wxSpline *spline)
{
    wxPoint *p;
    float           cx1, cy1, cx2, cy2, cx3, cy3, cx4, cy4;
    float           x1, y1, x2, y2;

    wxNode *node = spline->points->First();
    p = (wxPoint *)node->Data();

    x1 = p->x;
    y1 = p->y;

    node = node->Next();
    p = (wxPoint *)node->Data();

    x2 = p->x;
    y2 = p->y;
    cx1 = (float)((x1 + x2) / 2);
    cy1 = (float)((y1 + y2) / 2);
    cx2 = (float)((cx1 + x2) / 2);
    cy2 = (float)((cy1 + y2) / 2);

    wx_spline_add_point((int) x1, (int) y1);

    node = node->Next();

    while (node)
    {
        p = (wxPoint *)node->Data();
	x1 = x2;
	y1 = y2;
	x2 = p->x;
	y2 = p->y;
        cx4 = (float)(x1 + x2) / 2;
        cy4 = (float)(y1 + y2) / 2;
        cx3 = (float)(x1 + cx4) / 2;
        cy3 = (float)(y1 + cy4) / 2;

        wx_quadratic_spline(cx1, cy1, cx2, cy2, cx3, cy3, cx4, cy4);

	cx1 = cx4;
	cy1 = cy4;
        cx2 = (float)(cx1 + x2) / 2;
        cy2 = (float)(cy1 + y2) / 2;

        node = node->Next();
    }

    wx_spline_add_point(wx_round(cx1), wx_round(cy1));
    wx_spline_add_point((int) x2, (int) y2);

    wx_spline_draw_point_array(dc);

}

void wx_draw_open_spline_ps(wxPostScriptDC *dc, wxSpline *s)
{
	double		a, b, c, d, x1, y1, x2, y2, x3, y3;
        wxPoint *p, *q;
        dc->SetPen(dc->current_pen);

        wxNode *node = s->points->First();
        p = (wxPoint *)node->Data();

	x1 = p->x; y1 = p->y;

        node = node->Next();
        p = (wxPoint *)node->Data();
	c = p->x; d = p->y;
        x3 = a = (float)(x1 + c) / 2;
        y3 = b = (float)(y1 + d) / 2;

        *(dc->pstream) << "newpath " << x1 << " " << dc->yorigin - y1 << " moveto " << x3 << " " << dc->yorigin - y3;
        *(dc->pstream) << " lineto\n";
        dc->CalcBoundingBox((float)x1, (float)(dc->yorigin - y1));
        dc->CalcBoundingBox((float)x3, (float)(dc->yorigin - y3));

        node = node->Next();

        while (node)
	{
          q = (wxPoint *)node->Data();

	  x1 = x3; y1 = y3;
	  x2 = c;  y2 = d;
	  c = q->x; d = q->y;
          x3 = (float)(x2 + c) / 2;
          y3 = (float)(y2 + d) / 2;
          *(dc->pstream) << x1 << " " << dc->yorigin - y1 << " " << x2 << " " << dc->yorigin - y2 << " ";
          *(dc->pstream) << x3 << " " << dc->yorigin - y3 << " DrawSplineSection\n";

          dc->CalcBoundingBox((float)x1, (float)(dc->yorigin - y1));
          dc->CalcBoundingBox((float)x3, (float)(dc->yorigin - y3));


         node = node->Next();
        }
	/*
	* At this point, (x2,y2) and (c,d) are the position of the 
	* next-to-last and last point respectively, in the point list
	*/
        *(dc->pstream) << c << " " << dc->yorigin - d << " lineto stroke\n";
}

/********************* CURVES FOR SPLINES *****************************

	The following spline drawing routine is from

	"An Algorithm for High-Speed Curve Generation"
	by George Merrill Chaikin,
	Computer Graphics and Image Processing, 3, Academic Press,
	1974, 346-349.

	and

	"On Chaikin's Algorithm" by R. F. Riesenfeld,
	Computer Graphics and Image Processing, 4, Academic Press,
	1975, 304-310.

***********************************************************************/

#define		half(z1, z2)	((z1+z2)/2.0)
#define		THRESHOLD	5

/* iterative version */

void wx_quadratic_spline(float a1, float b1, float a2, float b2, float a3, float b3, float a4,
                 float b4)
{
    register float  xmid, ymid;
    float           x1, y1, x2, y2, x3, y3, x4, y4;

    wx_clear_stack();
    wx_spline_push(a1, b1, a2, b2, a3, b3, a4, b4);

    while (wx_spline_pop(&x1, &y1, &x2, &y2, &x3, &y3, &x4, &y4)) {
        xmid = (float)half(x2, x3);
        ymid = (float)half(y2, y3);
	if (fabs(x1 - xmid) < THRESHOLD && fabs(y1 - ymid) < THRESHOLD &&
	    fabs(xmid - x4) < THRESHOLD && fabs(ymid - y4) < THRESHOLD) {
            wx_spline_add_point(wx_round(x1), wx_round(y1));
            wx_spline_add_point(wx_round(xmid), wx_round(ymid));
	} else {
            wx_spline_push(xmid, ymid, (float)half(xmid, x3), (float)half(ymid, y3),
                 (float)half(x3, x4), (float)half(y3, y4), x4, y4);
            wx_spline_push(x1, y1, (float)half(x1, x2), (float)half(y1, y2),
                 (float)half(x2, xmid), (float)half(y2, ymid), xmid, ymid);
	}
    }
}


/* utilities used by spline drawing routines */


typedef struct wx_spline_stack_struct {
    float           x1, y1, x2, y2, x3, y3, x4, y4;
}
                Stack;

#define         SPLINE_STACK_DEPTH             20
static Stack    wx_spline_stack[SPLINE_STACK_DEPTH];
static Stack   *wx_stack_top;
static int      wx_stack_count;

void wx_clear_stack(void)
{
    wx_stack_top = wx_spline_stack;
    wx_stack_count = 0;
}

void wx_spline_push(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4)
{
    wx_stack_top->x1 = x1;
    wx_stack_top->y1 = y1;
    wx_stack_top->x2 = x2;
    wx_stack_top->y2 = y2;
    wx_stack_top->x3 = x3;
    wx_stack_top->y3 = y3;
    wx_stack_top->x4 = x4;
    wx_stack_top->y4 = y4;
    wx_stack_top++;
    wx_stack_count++;
}

int wx_spline_pop(float *x1, float *y1, float *x2, float *y2,
                  float *x3, float *y3, float *x4, float *y4)
{
    if (wx_stack_count == 0)
	return (0);
    wx_stack_top--;
    wx_stack_count--;
    *x1 = wx_stack_top->x1;
    *y1 = wx_stack_top->y1;
    *x2 = wx_stack_top->x2;
    *y2 = wx_stack_top->y2;
    *x3 = wx_stack_top->x3;
    *y3 = wx_stack_top->y3;
    *x4 = wx_stack_top->x4;
    *y4 = wx_stack_top->y4;
    return (1);
}

static Bool wx_spline_add_point(int x, int y)
{
  wxPoint *point = new wxPoint((float)x, (float)y);
  wx_spline_point_list.Append(point);
  return TRUE;
}

static void wx_spline_draw_point_array(wxbDC *dc)
{
  dc->DrawLines(&wx_spline_point_list, 0.0, 0.0);
  wxNode *node = wx_spline_point_list.First();
  while (node)
  {
    wxPoint *point = (wxPoint *)node->Data();
    delete point;
    delete node;
    node = wx_spline_point_list.First();
  }
}

wxSpline::wxSpline(wxList *list)
{
  points = list;
}

wxSpline::~wxSpline(void)
{
}

void wxSpline::DeletePoints(void)
{
  wxNode *node = points->First();
  while (node)
  {
    wxPoint *point = (wxPoint *)node->Data();
    delete point;
    delete node;
    node = points->First();
  }
  delete points;
}
