/*
 * File:     wx_ps.cc
 * Purpose:  Device context implementation (PostScript)
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Uncomment for Borland/UNIX combination
#ifdef wx_msw
#include "wx.h"
#pragma hdrstop
#else

#include "common.h"
#include "wx_frame.h"
#include "wx_dcps.h"
#include "wx_utils.h"
#include "wx_canvs.h"
#include "wx_dialg.h"
#include "wx_main.h"

#include "wx_lbox.h"
#include "wx_buttn.h"
#include "wx_choic.h"
#include "wx_check.h"
#include "wx_messg.h"
#include "wx_txt.h"
#include "wx_mtxt.h"
#endif
*/

#include "wx.h"                 // MSC 7/UNIX combination
#include "wx_privt.h"

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// Declarations local to this file
#define YSCALE(y) (yorigin - (y))
#define     wx_round(a)    (int)((a)+.5)

extern void wx_draw_open_spline_ps(wxPostScriptDC *dc, wxSpline *s);

#ifdef wx_x
Bool XPrinterDialog(void);
#endif

wxPostScriptDC::wxPostScriptDC(wxPostScriptDC *old_dc)
{
  __type = wxTYPE_DC_POSTSCRIPT;
  device = wxDEVICE_EPS;
  min_x = 0; min_y = 0; max_x = 0; max_y = 0;

  ok = FALSE;
  pstream = NULL;
  title = NULL;
  filename = NULL;

  current_logical_function = -1;
  font = NULL;
  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;

  ok = TRUE;

  Colour = wxColourDisplay();
  SetBrush(wxWHITE_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxPostScriptDC::wxPostScriptDC(char *file, Bool interactive)
{
  __type = wxTYPE_DC_POSTSCRIPT;
  wx_interactive = interactive;
  font = NULL;
  device = wxDEVICE_EPS;

  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  yorigin = 792;  // For EPS output
  min_x = 1000.0;
  min_y = 1000.0;
  max_x = -1000.0;
  max_y = -1000.0;
  title = NULL;
  if (file)
    filename = copystring(file);
  else filename = NULL;

  pstream = NULL;

#ifdef wx_x
  if (interactive)
    ok = XPrinterDialog();
  else ok = TRUE;
  if (!ok)
    return;

  if (interactive && (wx_preview || (!file && !wx_print_to_file)))
  {
    char userId[256];
    wxGetUserId(userId, sizeof(userId));
    strcpy(wx_printer_file, "/tmp/preview_");
    strcat(wx_printer_file, userId);
    strcat(wx_printer_file, ".ps");
    file = wx_printer_file;
  }
#endif

  if (device == wxDEVICE_EPS)
  {
    if (!file)
    {
      file = wxFileSelector("Enter PostScript file name", NULL, "", "ps", "*.ps");
      if (!file)
      {
        ok = FALSE;
        return;
      }
      strcpy(wx_printer_file, file);
      ok = TRUE;
    }
    else strcpy(wx_printer_file, file);

    pstream = new ofstream(wx_printer_file);
    if (!pstream)
    {
      (void)wxMessageBox("Error", "Cannot open file!", wxOK);
      ok = FALSE;
      return;
    }
    ok = TRUE;
  }
  current_logical_function = -1;
  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;
  Colour = wxColourDisplay();

  SetBrush(wxBLACK_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxPostScriptDC::~wxPostScriptDC(void)
{
  if (pstream)
    delete pstream;
  if (filename)
    delete[] filename;
}

void wxPostScriptDC::SetClippingRegion(float cx, float cy, float cw, float ch)
{
  if (clipping)
    return;

  clipping = TRUE;
  *pstream << "gsave\n";
  *pstream << "newpath\n";
  *pstream << cx << " " << YSCALE(cy) << " moveto\n";
  *pstream << cx+cw << " " << YSCALE(cy) << " lineto\n";
  *pstream << cx+cw << " " << YSCALE(cy + ch) << " lineto\n";
  *pstream << cx << " " << YSCALE(cy + ch) << " lineto\n";
  *pstream << "closepath clip newpath\n";
}

void wxPostScriptDC::DestroyClippingRegion(void)
{
  if (clipping)
  {
    clipping = FALSE;
    *pstream << "grestore\n";
  }
}

void wxPostScriptDC::Clear(void)
{
}

void wxPostScriptDC::DrawLine(float x1, float y1, float x2, float y2)
{
  if (current_pen) SetPen(current_pen);
  *pstream << "newpath\n";
  *pstream << x1 << " " << YSCALE(y1) << " moveto\n";
  *pstream << x2 << " " << YSCALE(y2) << " lineto\n";
  *pstream << "stroke\n";
  CalcBoundingBox(x1, YSCALE(y1));
  CalcBoundingBox(x2, YSCALE(y2));
}

void wxPostScriptDC::DrawArc(float x1,float y1,float x2,float y2,float xc,float yc)
{
  if (current_pen) SetPen(current_pen);
}

void wxPostScriptDC::DrawPoint(float x, float y)
{
  if (current_pen) SetPen(current_pen);
  *pstream << "newpath\n";
  *pstream << x << " " << YSCALE(y) << " moveto\n";
  *pstream << x << " " << YSCALE(y) << " lineto\n";
  *pstream << "stroke\n";
  CalcBoundingBox(x, YSCALE(y));
}

void wxPostScriptDC::DrawPolygon(int n, wxPoint points[], float xoffset, float yoffset)
{
  if (n > 0)
  {
    if (current_brush && current_brush->style != wxTRANSPARENT)
    {
      SetBrush(current_brush);
      *pstream << "newpath\n";

      float xx = points[0].x + xoffset;
      float yy = YSCALE(points[0].y + yoffset);
      *pstream << xx << " " << yy << " moveto\n";
      CalcBoundingBox(xx, yy);

      int i;
      for (i = 1; i < n; i++)
      {
        xx = points[i].x + xoffset;
        yy = YSCALE(points[i].y + yoffset);
        *pstream << xx << " " << yy << " lineto\n";
        CalcBoundingBox(xx, yy);
      }
      *pstream << "fill\n";
    }

    if (current_pen && current_pen->style != wxTRANSPARENT)
    {
      SetPen(current_pen);
      *pstream << "newpath\n";

      float xx = points[0].x + xoffset;
      float yy = YSCALE(points[0].y + yoffset);
      *pstream << xx << " " << yy << " moveto\n";
      CalcBoundingBox(xx, yy);

      int i;
      for (i = 1; i < n; i++)
      {
        xx = points[i].x + xoffset;
        yy = YSCALE(points[i].y + yoffset);
        *pstream << xx << " " << yy << " lineto\n";
        CalcBoundingBox(xx, yy);
      }

      // Close the polygon
      xx = points[0].x + xoffset;
      yy = YSCALE(points[0].y + yoffset);
      *pstream << xx << " " << yy << " lineto\n";

      // Output the line
      *pstream << "stroke\n";
    }
  }
}

void wxPostScriptDC::DrawLines(int n, wxPoint points[], float xoffset, float yoffset)
{
  if (n > 0)
  {
    if (current_pen) SetPen(current_pen);

    *pstream << "newpath\n";

    float xx = points[0].x + xoffset;
    float yy = YSCALE(points[0].y + yoffset);
    *pstream << xx << " " << yy << " moveto\n";
    CalcBoundingBox(xx, yy);

    int i;
    for (i = 1; i < n; i++)
    {
      xx = points[i].x + xoffset;
      yy = YSCALE(points[i].y + yoffset);
      *pstream << xx << " " << yy << " lineto\n";
      CalcBoundingBox(xx, yy);
    }
    *pstream << "stroke\n";
  }
}

void wxPostScriptDC::DrawRectangle(float x, float y, float width, float height)
{
  if (current_brush && current_brush->style != wxTRANSPARENT)
  {
    SetBrush(current_brush);
  
    *pstream << "newpath\n";
    *pstream << x << " " << YSCALE(y) << " moveto\n";
    *pstream << x+width << " " << YSCALE(y) << " lineto\n";
    *pstream << x+width << " " << YSCALE(y+height) << " lineto\n";
    *pstream << x << " " << YSCALE(y+height) << " lineto\n";
    *pstream << "closepath\n";
    *pstream << "fill\n";

    CalcBoundingBox(x, YSCALE(y));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
  if (current_pen && current_pen->style != wxTRANSPARENT)
  {
    SetPen(current_pen);
  
    *pstream << "newpath\n";
    *pstream << x << " " << YSCALE(y) << " moveto\n";
    *pstream << x+width << " " << YSCALE(y) << " lineto\n";
    *pstream << x+width << " " << YSCALE(y+height) << " lineto\n";
    *pstream << x << " " << YSCALE(y+height) << " lineto\n";
    *pstream << "closepath\n";
    *pstream << "stroke\n";

    CalcBoundingBox(x, YSCALE(y));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
}

void wxPostScriptDC::DrawRoundedRectangle(float x, float y, float width, float height, float radius)
{
  if (current_brush && current_brush->style != wxTRANSPARENT)
  {
    SetBrush(current_brush);
    // Draw rectangle anticlockwise
    *pstream << "newpath\n";
    *pstream << x+radius << " " << YSCALE(y+radius) << " " << radius << " 90 180 arc\n";

    *pstream << x << " " << YSCALE(y + radius) << " moveto\n";

    *pstream << x+radius << " " << YSCALE(y+height-radius) << " " << radius << " 180 270 arc\n";
    *pstream << x+width-radius << " " << YSCALE(y+height) << " lineto\n";

    *pstream << x+width-radius << " " << YSCALE(y+height-radius) << " " << radius << " 270 0 arc\n";
    *pstream << x+width << " " << YSCALE(y+radius) << " lineto\n";

    *pstream << x+width-radius << " " << YSCALE(y+radius) << " " << radius << " 0 90 arc\n";

    *pstream << x+radius << " " << YSCALE(y) << " lineto\n";

    *pstream << "closepath\n";

    *pstream << "fill\n";

    CalcBoundingBox(x, YSCALE(y));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
  if (current_pen && current_pen->style != wxTRANSPARENT)
  {
    SetPen(current_pen);
    // Draw rectangle anticlockwise
    *pstream << "newpath\n";
    *pstream << x+radius << " " << YSCALE(y+radius) << " " << radius << " 90 180 arc\n";

    *pstream << x << " " << YSCALE(y + height - radius) << " lineto\n";

    *pstream << x+radius << " " << YSCALE(y+height-radius) << " " << radius << " 180 270 arc\n";
    *pstream << x+width-radius << " " << YSCALE(y+height) << " lineto\n";

    *pstream << x+width-radius << " " << YSCALE(y+height-radius) << " " << radius << " 270 0 arc\n";
    *pstream << x+width << " " << YSCALE(y+radius) << " lineto\n";

    *pstream << x+width-radius << " " << YSCALE(y+radius) << " " << radius << " 0 90 arc\n";

    *pstream << x+radius << " " << YSCALE(y) << " lineto\n";

    *pstream << "closepath\n";

    *pstream << "stroke\n";

    CalcBoundingBox(x, YSCALE(y));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
}

void wxPostScriptDC::DrawEllipse(float x, float y, float width, float height)
{
  if (current_brush && current_brush->style != wxTRANSPARENT)
  {
    SetBrush(current_brush);

    *pstream << "newpath\n";
    *pstream << x + width/2 << " " << YSCALE(y + height/2) << " ";
    *pstream << width/2 << " " << height/2 << " 0 360 ellipse\n";
    *pstream << "fill\n";

    CalcBoundingBox(x-width, YSCALE(y-height));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
  if (current_pen && current_pen->style != wxTRANSPARENT)
  {
    SetPen(current_pen);

    *pstream << "newpath\n";
    *pstream << x + width/2 << " " << YSCALE(y + height/2) << " ";
    *pstream << width/2 << " " << height/2 << " 0 360 ellipse\n";
    *pstream << "stroke\n";

    CalcBoundingBox(x-width, YSCALE(y-height));
    CalcBoundingBox(x+width, YSCALE(y+height));
  }
}

void wxPostScriptDC::DrawIcon(wxIcon *icon, float x, float y)
{
}


void wxPostScriptDC::SetFont(wxFont *the_font)
{
  font = the_font;
  char buf[100];
  char *name;
  char *style = "";
  int Style = font->style;
  int Weight = font->weight;

  switch (font->family)
  {
    case wxMODERN:     name = "/Courier";
           break;
    case wxSWISS:  name = "/Helvetica";
           break;
    default:
    case wxDEFAULT:
    case wxROMAN:  name = "/Times-Roman";
           break;
  }

  if (Style == wxNORMAL && (Weight == wxNORMAL || Weight == wxLIGHT))
  {
    if (font->family == wxROMAN)
      style = "-Roman";
    else
      style = "";
  }
  else if (Style == wxNORMAL && Weight == wxBOLD)
    style = "-Bold";

  else if (Style == wxITALIC && (Weight == wxNORMAL || Weight == wxLIGHT))
  {
    if (font->family == wxROMAN)
      style = "-Italic";
    else
      style = "-Oblique";
  }
  else if (Style == wxITALIC && Weight == wxBOLD)
  {
    if (font->family == wxROMAN)
      style = "-BoldItalic";
    else
      style = "-BoldOblique";
  }
  else if (Style == wxSLANT && (Weight == wxNORMAL || Weight == wxLIGHT))
  {
    if (font->family == wxROMAN)
      style = "-Italic";
    else
      style = "-Oblique";
  }
  else if (Style == wxSLANT && Weight == wxBOLD)
  {
    if (font->family == wxROMAN)
      style = "-BoldItalic";
    else
      style = "-BoldOblique";
  }
  else style = "";

  strcpy(buf, name);
  strcat(buf, style);
  *pstream << buf << " findfont\n";
  *pstream << font->point_size << " scalefont setfont\n";
}

void wxPostScriptDC::SetPen(wxPen *pen)
{
  current_pen = pen;
  if (!pen)
    return;
  // Line width
  *pstream << pen->width << " setlinewidth\n";

  // Line style - WRONG: 2nd arg is OFFSET
/*
  Here, I'm afraid you do not conceive meaning of parameters of 'setdash'
  operator correctly. You should look-up this in the Red Book: the 2nd parame-
  ter is not number of values in the array of the first one, but an offset
  into this description of the pattern. I mean a real *offset* not index
  into array. I.e. If the command is [3 4] 1 setdash   is used, then there
  will be first black line *2* units long, then space 4 units, then the
  pattern of *3* units black, 4 units space will be repeated.
*/
  static char *dotted = "[2 5] 2";
  static char *short_dashed = "[4 4] 2";
  static char *long_dashed = "[4 8] 2";
  static char *dotted_dashed = "[6 6 2 6] 4";

  char *psdash = NULL;
  switch (pen->style)
  {
    case wxDOT:
      psdash = dotted;
      break;
    case wxSHORT_DASH:
      psdash = short_dashed;
      break;
    case wxLONG_DASH:
      psdash = long_dashed;
      break;
    case wxDOT_DASH:
      psdash = dotted_dashed;
      break;
    case wxSOLID:
    case wxTRANSPARENT:
    default:
      psdash = "[] 0";
      break;
  }
  *pstream << psdash << " setdash\n";

  // Line colour
  unsigned char red = pen->colour->Red();
  unsigned char blue = pen->colour->Blue();
  unsigned char green = pen->colour->Green();

  if (!Colour)
  {
    // Anything not white is black
    if (!(red == (unsigned char )255 && blue == (unsigned char)255
           && green == (unsigned char)255))
    {
      red = (unsigned char )0;
      green = (unsigned char )0;
      blue = (unsigned char )0;
    }
  }
  float redPS = (float)(((int)red)/255.0);
  float bluePS = (float)(((int)blue)/255.0);
  float greenPS = (float)(((int)green)/255.0);
  *pstream << redPS << " " << greenPS << " " << bluePS << " setrgbcolor\n";
}

void wxPostScriptDC::SetBrush(wxBrush *brush)
{
  current_brush = brush;
  if (!brush)
  {
    return;
  }
  // Brush colour
  unsigned char red = brush->colour->Red();
  unsigned char blue = brush->colour->Blue();
  unsigned char green = brush->colour->Green();

  if (!Colour)
  {
    // Anything not black is white
    if (!(red == (unsigned char )0 && blue == (unsigned char)0
           && green == (unsigned char)0))
    {
      red = (unsigned char )255;
      green = (unsigned char )255;
      blue = (unsigned char )255;
    }
  }
  float redPS = (float)(((int)red)/255.0);
  float bluePS = (float)(((int)blue)/255.0);
  float greenPS = (float)(((int)green)/255.0);
  *pstream << redPS << " " << greenPS << " " << bluePS << " setrgbcolor\n";
}

void wxPostScriptDC::DrawText(char *text, float x, float y)
{
  if (font)
    SetFont(font);

  if (current_text_foreground.Ok())
  {
    unsigned char red = current_text_foreground.Red();
    unsigned char blue = current_text_foreground.Blue();
    unsigned char green = current_text_foreground.Green();

    if (!Colour)
    {
      // Anything not white is black
      if (!(red == (unsigned char )255 && blue == (unsigned char)255
         && green == (unsigned char)255))
      {
        red = (unsigned char )0;
        green = (unsigned char )0;
        blue = (unsigned char )0;
      }
    }
    float redPS = (float)(((int)red)/255.0);
    float bluePS = (float)(((int)blue)/255.0);
    float greenPS = (float)(((int)green)/255.0);
    *pstream << redPS << " " << greenPS << " " << bluePS << " setrgbcolor\n";
  }

  int size = 10;
  if (font) size = font->point_size;

  *pstream << x << " " << YSCALE(y + size) << " moveto\n";

//  *pstream << "(" << text << ")" << " show\n";
  *pstream << "(";
  int len = strlen(text);
  for (int i = 0; i < len; i++)
  {
    char ch = text[i];
    if (ch == ')' || ch == '(' || ch == '\\')
      *pstream << "\\";
    *pstream << ch;
  }

  *pstream << ")" << " show\n";

  CalcBoundingBox(x, YSCALE(y+size));
  CalcBoundingBox(x + size*strlen(text), YSCALE(y));
}

// Make a 3-point spline
void wxPostScriptDC::DrawSpline(float x1, float y1, float x2, float y2, float x3, float y3)
{
  wxList *point_list = new wxList;

  wxPoint *point1 = new wxPoint;
  point1->x = x1; point1->y = y1;
  point_list->Append(point1);

  wxPoint *point2 = new wxPoint;
  point2->x = x2; point2->y = y2;
  point_list->Append(point2);

  wxPoint *point3 = new wxPoint;
  point3->x = x3; point3->y = y3;
  point_list->Append(point3);

  wxSpline spline(point_list);

  wx_draw_open_spline_ps(this, &spline);
  spline.DeletePoints();
}

void wxPostScriptDC::DrawSpline(wxList *list)
{
  wxSpline spline(list);

  wx_draw_open_spline_ps(this, &spline);
}

void wxPostScriptDC::SetBackground(wxBrush *brush)
{
  current_background_brush = brush;
}

void wxPostScriptDC::SetLogicalFunction(int function)
{
}

#define ELLIPSE_PS  "\
/ellipsedict 8 dict def\n\
ellipsedict /mtrx matrix put\n\
/ellipse\n\
{ ellipsedict begin\n\
  /endangle exch def\n\
  /startangle exch def\n\
  /yrad exch def\n\
  /xrad exch def\n\
  /y exch def\n\
  /x exch def\n\
  /savematrix mtrx currentmatrix def\n\
  x y translate\n\
  xrad yrad scale\n\
  0 0 1 startangle endangle arc\n\
  savematrix setmatrix\n\
  end\n\
  } def\n\
"

#define		SPLINE_PS	" \
/DrawSplineSection {\n\
	/y3 exch def\n\
	/x3 exch def\n\
	/y2 exch def\n\
	/x2 exch def\n\
	/y1 exch def\n\
	/x1 exch def\n\
	/xa x1 x2 x1 sub 0.666667 mul add def\n\
	/ya y1 y2 y1 sub 0.666667 mul add def\n\
	/xb x3 x2 x3 sub 0.666667 mul add def\n\
	/yb y3 y2 y3 sub 0.666667 mul add def\n\
	x1 y1 lineto\n\
	xa ya xb yb x3 y3 curveto\n\
	} def\n\
"

Bool wxPostScriptDC::StartDoc(char *message)
{
  wxPageNumber = 1;
  if (message) title = copystring(message);
  return TRUE;
}

void wxPostScriptDC::EndDoc(void)
{
  if (clipping) { clipping = FALSE; *pstream << "grestore\n"; }

  // Will reuse pstream for header
  if (pstream) { delete pstream; pstream = NULL; }

#ifdef wx_x
  char buf[300];
  long when;
#endif
#ifdef wx_msw
  time_t when;
#endif
  (void) time(&when);

  // Write header now
#ifdef wx_x
  char *header_file = "/tmp/_header.ps";
#endif
#ifdef wx_msw
  char *header_file = "_header.ps";
#endif
  pstream = new ofstream(header_file);

  *pstream << "%!PS-Adobe-EPSF\n";   /* PostScript magic strings */
  if (title) *pstream << "%%Title: " << title << "\n";
  *pstream << "%%Creator: " << wxTheApp->argv[0] << "\n";
  *pstream << "%%CreationDate: " << ctime(&when);
#ifdef wx_x
  char host[256];
  char userId[256];
  char userName[256];
  wxGetHostName(host, sizeof(host));
  wxGetUserId(userId, sizeof(userId));
  wxGetUserName(userName, sizeof(userName));
  *pstream << "%%For: " << userId << "@" << host << " (" <<
    userName << ")\n";
#endif
  // In landscape mode, what should bounding box be changed to????
  // Bounding box is currently wrong in landscape mode.

  *pstream << "%%BoundingBox: " << min_x << " " << min_y << " " << max_x << " " << max_y << "\n";
  *pstream << "%%Pages: " << wxPageNumber - 1 << "\n";
  *pstream << "%%EndComments\n\n";

  // Output scaling
  float real_translate_y = wx_printer_translate_y;
  if (!wx_portrait)
  {
    real_translate_y -= max_y;
    *pstream << "90 rotate\n";
  }

  *pstream << wx_printer_scale_x << " " << wx_printer_scale_y << " scale\n";
  *pstream << wx_printer_translate_x << " " << real_translate_y << " translate\n";

  *pstream << ELLIPSE_PS;
  *pstream << SPLINE_PS;

  delete pstream;
  pstream = NULL;

#ifdef wx_x
  char *concat_name = "/tmp/_concat.tmp";
#endif
#ifdef wx_msw
  char *concat_name = "concat.tmp";
#endif

  // Would be better not to write to wx_printer_file initially,
  // clear up sometime
  wxConcatFiles(header_file, wx_printer_file, concat_name);
  wxRemoveFile(wx_printer_file);
  wxCopyFile(concat_name, wx_printer_file);
  wxRemoveFile(concat_name);
  wxRemoveFile(header_file);
#ifdef wx_x
  if (wx_interactive)
  {
    if (ok && wx_preview)
    {
      strcpy(buf, wx_preview_command);
      strcat(buf, " ");
      strcat(buf, wx_printer_file);
      strcat(buf, " &");
      system(buf);
    }
    else if (ok && !wx_preview && !wx_print_to_file)
    {
      strcpy(buf, wx_printer_command);
      strcat(buf, " ");
      strcat(buf, wx_printer_flags);
      strcat(buf, " ");
      strcat(buf, wx_printer_file);
      strcat(buf, " &");
      system(buf);
    }
  }
#endif
}

void wxPostScriptDC::StartPage(void)
{
  *pstream << "%%Page: " << wxPageNumber ++ << "\n";
}

void wxPostScriptDC::EndPage(void)
{
  *pstream << "showpage\n";
}

Bool wxPostScriptDC::Blit(float xdest, float ydest, float width, float height,
                wxCanvasDC *source, float xsrc, float ysrc, int rop)
{
  return FALSE;
}

float wxPostScriptDC::GetCharHeight(void)
{
      if (font)
        return (float)font->point_size;
      else return 12.0;
}


float wxPostScriptDC::GetCharWidth(void)
{
      return 0;
}

void wxPostScriptDC::GetTextExtent(char *string, float *x, float *y,
                                   float *descent, float *externalLeading)
{
      // Provide a VERY rough estimate (avoid using it)
      int width = 12;
      int height = 12;

      if (font)
        { height = font->point_size; width = height; }
      *x = (float)strlen(string)*width;
      *y = (float)height;
      if (descent) *descent = 0.0;
      if (externalLeading) *externalLeading = 0.0;
}

void wxPostScriptDC::SetMapMode(int mode)
{
  mapping_mode = mode;
  return;
}

void wxPostScriptDC::SetUserScale(float x, float y)
{
  user_scale_x = x;
  user_scale_y = y;
}

float wxPostScriptDC::DeviceToLogicalX(int x)
{
  return (float)x;
}

float wxPostScriptDC::DeviceToLogicalXRel(int x)
{
  return (float)x;
}

float wxPostScriptDC::DeviceToLogicalY(int y)
{
  return (float)y;
}

float wxPostScriptDC::DeviceToLogicalYRel(int y)
{
  return (float)y;
}

int wxPostScriptDC::LogicalToDeviceX(float x)
{
  return (int)x;
}

int wxPostScriptDC::LogicalToDeviceXRel(float x)
{
  return (int)x;
}

int wxPostScriptDC::LogicalToDeviceY(float y)
{
  return (int)y;
}

int wxPostScriptDC::LogicalToDeviceYRel(float y)
{
  return (int)y;
}

class wxPrinterDialogBox: public wxDialogBox
{
 public:
  wxPrinterDialogBox(wxFrame *frame, char *title, Bool modal = FALSE,
              int x = -1, int y = -1, int
              width = -1, int height = -1);
};

#ifdef wx_x
wxPrinterDialogBox::wxPrinterDialogBox(wxFrame *frame, char *title, Bool modal,
              int x, int y, int width, int height):
  wxDialogBox(frame, title, modal, x, y, width, height)
{
}

Bool wxPrinterDialogAnswer = TRUE;

void wxPrinterDialogOk(wxButton& button, wxEvent& event)
{
  wxPrinterDialogAnswer = TRUE;
  wxPrinterDialogBox *dialog = (wxPrinterDialogBox *)button.GetParent();
  dialog->Show(FALSE);
}

void wxPrinterDialogCancel(wxButton& button, wxEvent& event)
{
  wxPrinterDialogAnswer = FALSE;
  wxPrinterDialogBox *dialog = (wxPrinterDialogBox *)button.GetParent();
  dialog->Show(FALSE);
}

Bool XPrinterDialog(void)
{
  char buf[100];
  wxPrinterDialogBox dialog(NULL, "Printer Settings", TRUE, 150, 150, 400, 400);

  (void)new wxButton(&dialog, (wxFunction)wxPrinterDialogOk, "Ok");
  (void)new wxButton(&dialog, (wxFunction)wxPrinterDialogCancel, "Cancel");
  dialog.NewLine();
  dialog.NewLine();

  wxText text_prt(&dialog, (wxFunction)NULL, "Printer Command: ", wx_printer_command, -1, -1, 100, -1);
  dialog.NewLine();

  wxText text0(&dialog, (wxFunction)NULL, "Printer Options: ", wx_printer_flags, -1, -1, 200, -1);
  dialog.NewLine();

  wxCheckBox checkbox0(&dialog, (wxFunction)NULL, "Portrait");
  checkbox0.SetValue(wx_portrait);

  wxCheckBox checkbox1(&dialog, (wxFunction)NULL, "Print to file");
  checkbox1.SetValue(wx_print_to_file);

  wxCheckBox checkbox2(&dialog, (wxFunction)NULL, "Preview only");
  checkbox2.SetValue(wx_preview);
  dialog.NewLine();

  sprintf(buf, "%.2f", wx_printer_scale_x);
  wxText text1(&dialog, (wxFunction)NULL, "X Scaling: ", buf, -1, -1, 100, -1);

  sprintf(buf, "%.2f", wx_printer_scale_y);
  wxText text2(&dialog, (wxFunction)NULL, "Y Scaling: ", buf, -1, -1, 100, -1);

  dialog.NewLine();

  sprintf(buf, "%.2f", wx_printer_translate_x);
  wxText text3(&dialog, (wxFunction)NULL, "X Translation: ", buf, -1, -1, 100, -1);

  sprintf(buf, "%.2f", wx_printer_translate_y);
  wxText text4(&dialog, (wxFunction)NULL, "Y Translation: ", buf, -1, -1, 100, -1);

  dialog.NewLine();

  dialog.Fit();

  dialog.Show(TRUE);

  if (wxPrinterDialogAnswer)
  {
    StringToFloat(text1.GetValue(), &wx_printer_scale_x);
    StringToFloat(text2.GetValue(), &wx_printer_scale_y);
    StringToFloat(text3.GetValue(), &wx_printer_translate_x);
    StringToFloat(text4.GetValue(), &wx_printer_translate_y);

    wx_printer_flags = copystring(text0.GetValue());
    wx_printer_command = copystring(text_prt.GetValue());

    wx_portrait = checkbox0.GetValue();
    wx_print_to_file = checkbox1.GetValue();
    wx_preview = checkbox2.GetValue();
  }

  return wxPrinterDialogAnswer;
}
#endif
