/*
 * File:     wx_sysev.cc
 * Purpose:  System event implementation
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Uncomment for Borland/UNIX combination
#ifdef wx_msw
#include "wx.h"
#pragma hdrstop
#else

#include "common.h"
#include "wx_utils.h"
#include "wx_list.h"
#endif
*/

#include "wx.h"                 // MSC 7/UNIX

#include "wx_sysev.h"

wxList wxEventClassList(wxKEY_INTEGER);
wxList wxEventNameList(wxKEY_INTEGER);
wxList wxPrimaryEventHandlerList;
wxList wxPreEventHandlerList;
wxList wxPostEventHandlerList;

class wxEventHandlerStruc: public wxObject
{
 public:
   wxEventHandler handler;
   Bool override;
};

/*
 * A different kind of event from wxEvent: general wxWindows events, covering
 * all interesting things that might happen (button clicking, resizing,
 * setting text in widgets, etc.).
 *
 * For each completely new event type, derive a new event class.
 *
 */
 
wxEvent::wxEvent(void)
{
  eventClass = 0;
  eventType = 0;
  objectType = 0;
  eventObject = NULL;
  eventHandle = NULL;
}

wxEvent::~wxEvent(void)
{
}


Bool wxSendEvent1(wxEvent& event, Bool external, wxList& list, Bool *stop)
{
  Bool handled = FALSE;
  wxNode *node = list.First();
  while (node)
  {
    wxEventHandlerStruc *thing = (wxEventHandlerStruc *)node->Data();
    handled = (*thing->handler)(&event, external);
    if (handled)
    {
      if (thing->override)
      {
        *stop = TRUE;
        return TRUE;
      } else *stop = FALSE;
    }
    node = node->Next();
  }
  return handled;
}

// Send an event to the system, usually will be external, but specify
// external = FALSE if calling from within the main application in response
// to other events.
// Returns TRUE iff the event was processed.
Bool wxSendEvent(wxEvent& event, Bool external)
{
  Bool stop = FALSE;
  return wxSendEvent1(event, external, wxPrimaryEventHandlerList, &stop);
}

// Notify the system of the event you are about to execute/have just executed.
// If TRUE is returned and pre = TRUE, calling code should not execute the event
// (i.e. it's been intercepted by a handler and vetoed).
// These events are always internal, because they're generated from within
// the main application code.
Bool wxNotifyEvent(wxEvent& event, Bool pre)
{
  Bool stop = FALSE;
  wxList *theList = NULL;
  if (pre)
    theList = &wxPreEventHandlerList;
  else
    theList = &wxPostEventHandlerList;

  return wxSendEvent1(event, FALSE, *theList, &stop);
}

// Add a primary event handler. pre = TRUE iff it should be called before the
// Returns TRUE if succeeds.
Bool wxAddPrimaryEventHandler(wxEventHandler handlerFunc)
{
  wxEventHandlerStruc *struc = new wxEventHandlerStruc;
  struc->handler = handlerFunc;
  struc->override = FALSE;

  wxPrimaryEventHandlerList.Append(struc);
  return TRUE;
}

// Add a secondary event handler, pre = TRUE iff it should be called before the
// event is executed. override = TRUE iff the handler is allowed to override
// all subsequent events by returning TRUE.
// Returns TRUE if succeeds.
Bool wxAddSecondaryEventHandler(wxEventHandler handlerFunc,
                                Bool pre, Bool override, Bool append)
{
  wxEventHandlerStruc *struc = new wxEventHandlerStruc;
  struc->handler = handlerFunc;
  struc->override = override;

  if (pre)
  {
    if (append)
      wxPreEventHandlerList.Append(struc);
    else
      wxPreEventHandlerList.Insert(struc);
  }
  else
  {
    if (append)
      wxPostEventHandlerList.Append(struc);
    else
      wxPostEventHandlerList.Insert(struc);
  }

  return TRUE;
}

// Remove secondary event handler. Returns TRUE if succeeds.
Bool wxRemoveSecondaryEventHandler(wxEventHandler handlerFunc, Bool pre)
{
  wxList *theList;
  if (pre)
    theList = &wxPreEventHandlerList;
  else
    theList = &wxPostEventHandlerList;
  
  wxNode *node = theList->First();
  while (node)
  {
    wxEventHandlerStruc *thing = (wxEventHandlerStruc *)node->Data();
    if (thing->handler == handlerFunc)
    {
      delete thing;
      delete node;
      return TRUE;
    }
    else node = node->Next();
  }
  return FALSE;
}

// Register a new event class (derived from wxEvent),
// giving the new event class type, its superclass, a function for creating
// a new event object of this class, and an optional description.
void wxRegisterEventClass(WXTYPE eventClassId, WXTYPE superClassId,
                          wxEventConstructor constructor, char *description)
{
  wxSystemEventClassStruc *struc = new wxSystemEventClassStruc;
  struc->eventClass = eventClassId;
  struc->eventConstructor = constructor;
  if (description)
    struc->eventDescription = copystring(description);
  else
    struc->eventDescription = NULL;

  wxEventClassList.Append((long)eventClassId, struc);

//  wxAllTypes.AddType(eventClassId, superClassId, description);
}

// Register the name of the event. This will allow a simple command language
// where giving the event type name and some arguments will cause
// a new event of class eventClassId to be created, with given event type,
// and some arguments, allows an event to be dynamically constructed and sent.
void wxRegisterEventName(WXTYPE eventTypeId, WXTYPE eventClassId, char *eventName)
{
  wxSystemEventNameStruc *struc = new wxSystemEventNameStruc;
  struc->eventClass = eventClassId;
  struc->eventType = eventTypeId;
  struc->eventName = copystring(eventName);

  wxEventNameList.Append((long)eventTypeId, struc);
}

// Define this and link before wxWindows library to allow
// registering events from 'outside' the main application
void wxRegisterExternalEventHandlers(void)
{
}

/*
 * Event reading/writing helper functions
 *
 */

void wxWriteString(ostream& out, char *s)
{
  out << '"';

  int len = strlen(s);
  for (int i = 0; i < len; i++)
    if (s[i] == '"')
      out << '\\' << '"';
    else out << s[i];

  out << '"';
}

void wxWriteInteger(ostream& out, int i)
{
  out << i;
}

void wxWriteLong(ostream& out, long i)
{
  out << i;
}

void wxWriteFloat(ostream& out, float f)
{
  out << f;
}

void wxWriteSpace(ostream& out)
{
  out << " ";
}

Bool wxReadWhiteSpace(istream& in)
{
// Why is there no eatwhite in GCC?
#ifdef __MSC
  in.eatwhite();
#else
  char ch = in.peek();
  while ((ch == ' ') || (ch == '\n') || (ch == '\t'))
  {
    in.get();
    ch = in.peek();
  }
#endif
  return TRUE;
}

extern char wxBuffer[];
Bool wxReadString(istream& in, char **s)
{
  wxReadWhiteSpace(in);

  if (in.peek() != '"')
    return FALSE;
  in.get();
  Bool flag = TRUE;
  char ch = 0;
  int i = 0;
  while (flag && (i < 1000))
  {
    ch = in.get();
    switch (ch)
    {
      case EOF:
      {
        wxBuffer[i] = 0;
        flag = FALSE;
        break;
      }
      case '"':
      {
        wxBuffer[i] = 0;
        flag = FALSE;
        break;
      }
      case '\\':
      {
        if (in.peek() == '"')
        {
          in.get();
          wxBuffer[i] = '"';
        }
        else wxBuffer[i] = '\\';
        break;
      }
      default:
      {
        wxBuffer[i] = ch;
      }
    }
    ch = in.get();
    i ++;
  }
  wxBuffer[i] = 0;
  *s = copystring(wxBuffer);
  return TRUE;
}

Bool wxReadInteger(istream& in, int *theInt)
{
  in >> *theInt;
  return TRUE;
}

Bool wxReadLong(istream& in, long *theLong)
{
  in >> *theLong;
  return TRUE;
}

Bool wxReadFloat(istream& in, float *theFloat)
{
  in >> *theFloat;
  return TRUE;
}

