/*
 * File:     wx_enhdg.cc
 * Purpose:  wxEnhancedDialogBox
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Uncomment for Borland/UNIX combination
#ifdef wx_msw
#include "wx.h"
#pragma hdrstop
#else

#include "common.h"
#include "wx_dialg.h"
#include "wx_utils.h"
#include "wx_frame.h"
#include "wx_panel.h"
#include "wx_item.h"
#include "wx_buttn.h"
#include "wx_messg.h"
#include "wx_choic.h"
#include "wx_check.h"
#include "wx_menu.h"
#include "wx_txt.h"
#include "wx_mtxt.h"
#include "wx_slidr.h"
#include "wx_lbox.h"
#include "wx_enhdg.h"
#endif
*/

#include "wx.h"                 // MSC 7/UNIX
#include <iostream.h>
#include <stdio.h>

//
//	Enhanced Dialog Box
//
#ifndef wx_xview
static        wxFont  *pin_font ;
#endif
static        wxFont  *label_font ;
static        wxFont  *button_font ;
static        wxFont  *cmd_font ;

#ifndef wx_xview
void UnshowEnhDialog(wxCheckBox &pin,wxCommandEvent evt)
{
  if (pin.GetValue())
    return ;
  wxPanel *pinpanel = (wxPanel*)pin.window_parent ;
  wxEnhDialogBox *dialog = (wxEnhDialogBox*)pinpanel->window_parent ;
  dialog->Show(FALSE) ;
}
#endif

wxEnhDialogBox::wxEnhDialogBox(wxFrame *Parent, char *Title, Bool Modal, 
                         wxFunction fun,int space,
                         int x, int y, int width, int height,
			 char *name)
		: wxDialogBox(Parent,Title,Modal,x,y,width,height,wxXVIEW_CMD|DEFAULT_DIALOG_STYLE,name) 
{
  __type = wxTYPE_ENHANCED_DIALOG ;
  // Put your own defaults fonts here!
#ifdef ENHANCED_FONTS
#ifndef wx_xview
  // Under XView, font setting doesn't works. 2 possibles reasons:
  // 1) I make a mistake
  // 2) XView doesn't allow font setting
  // Until clear, no font changes for XView!!
  if (label_font==NULL)
    label_font = new wxFont(12,wxSWISS,wxNORMAL,wxBOLD) ;
  if (button_font==NULL)
    button_font = new wxFont(12,wxSWISS,wxNORMAL,wxNORMAL) ;
  if (pin_font==NULL)
    pin_font = new wxFont(14,wxROMAN,wxITALIC,wxBOLD) ;
  if (cmd_font==NULL)
    cmd_font = new wxFont(17,wxROMAN,wxNORMAL,wxNORMAL) ;
#endif
#endif
  panel = this ;
  panel->SetVerticalSpacing(0) ;
  panel->SetHorizontalSpacing(0) ;
  panel->SetLabelFont(label_font) ;
  panel->SetButtonFont(button_font) ;
#ifndef wx_xview
  pinPanel = new wxPanel(panel) ;
  pinPanel->SetVerticalSpacing(0) ;
  pinPanel->SetButtonFont(pin_font) ;
#ifdef wx_motif
  // Motif provide enough spacing...
  pinPanel->SetHorizontalSpacing(0) ;
#endif
  pinCheck = new wxCheckBox(pinPanel,(wxFunction)UnshowEnhDialog,"Pinned",
                            -1,-1,100,20,0,"pin") ;
  pinPanel->Fit() ;
  panel->NewLine() ;
#endif

  unshow = fun ;
  userPanel = new wxPanel(panel,-1,-1,-1,-1,0,"user") ;
  cmdPanel = 0 ;
  statusPanel = 0 ;
  statusText = 0 ;
  SetPin(FALSE) ;
  userSpacing = space ;
}

wxEnhDialogBox::~wxEnhDialogBox(void)
{
}

wxButton* wxEnhDialogBox::AddCmd(char *label,wxFunction fun,int tag)
{
  if (cmdPanel==0)
  {
#if 0
int ww,hh;
userPanel->GetSize(&ww,&hh) ;
fprintf(stderr,"Before Fit Userpanel %dx%d\n",ww,hh) ;
#endif
    userPanel->Fit() ;
#if 0
userPanel->GetSize(&ww,&hh) ;
fprintf(stderr,"After Fit Userpanel %dx%d\n",ww,hh) ;
#endif
    panel->NewLine() ;
    cmdPanel = new wxPanel(panel) ;
    cmdPanel->SetVerticalSpacing(0) ;
    cmdPanel->SetButtonFont(cmd_font) ;
    if (userSpacing>-1)
      cmdPanel->SetHorizontalSpacing(userSpacing) ;
  }

  wxButton *btn = new wxButton(cmdPanel,fun,label) ;
  if (btn)
    btn->SetClientData((char*)tag) ;
  return(btn) ;
}

wxButton* wxEnhDialogBox::GetCmd(int number)
{
  if (cmdPanel)
    return ((wxButton*)(cmdPanel->GetChild(number)));
  else
    return(NULL) ;
}

void wxEnhDialogBox::Show(Bool show,Bool flag)
{

  if (!show)
  {
#ifndef wx_xview
    if (pinCheck->GetValue())
      return ;
#endif
#ifdef wx_xview
    wxFrame *frame = (wxFrame*)GetParent() ;
    Frame x_frame = (Frame)(frame->GetHandle());
    if (xv_get(x_frame,FRAME_CMD_PUSHPIN_IN))
      return ;
#endif
    if (unshow)
    {
    wxCommandEvent evt = 0;
      (void)(*(unshow))(*this,evt) ;
    }
  }
  if (show)
    SetPin(flag) ;
  wxDialogBox::Show(show) ;
}

void wxEnhDialogBox::SetPin(Bool flag)
{
#ifndef wx_xview
  if (flag)
  {
    if (pinCheck->GetValue())
      return ;
  }
  else
  {
    if (!pinCheck->GetValue())
      return ;
  }
  // Change state only if needed
  pinCheck->SetValue(flag) ;
#endif
#ifdef wx_xview
    wxFrame *frame = (wxFrame*)GetParent() ;
    Frame x_frame = (Frame)(frame->GetHandle());
    (void)xv_set(x_frame,FRAME_CMD_PUSHPIN_IN,flag) ;
#endif
}

void wxEnhDialogBox::SetStatus(char *label)
{
  if (label==0 || *label=='\0')
    label = "   " ;
  statusText->SetLabel(label) ;
}

void wxEnhDialogBox::Fit(void)
{
  if (cmdPanel)
    cmdPanel->Fit() ;
  panel->NewLine() ;
  statusPanel = new wxPanel(panel) ;
  statusPanel->SetVerticalSpacing(0) ;
#ifdef wx_motif
  // Motif provide enough spacing...
  statusPanel->SetHorizontalSpacing(0) ;
#endif
  statusText = new wxMessage(statusPanel,"   ") ;
  statusPanel->Fit() ;
  panel->NewLine() ;
  wxDialogBox::Fit() ;
  if (cmdPanel)
    cmdPanel->Centre(wxHORIZONTAL) ;
  int fw,fh ;
  GetClientSize(&fw,&fh) ;
#if 0
fprintf(stderr,"Final get client size %dx%d\n",fw,fh) ;
#endif
  statusPanel->SetSize(-1,-1,fw,-1) ;
  statusText->SetSize(-1,-1,fw,-1) ;
}
