/*
 * File:     wx_dc.cc
 * Purpose:  Device context implementation
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "wx.h"
#pragma hdrstop
#include "wx_privt.h"
#include "math.h"

#include <commdlg.h>
// #include <math.h>
// #include <stdio.h>
// #include <stdlib.h>
// #include <time.h>

/*
#include "common.h"
#include "wx_frame.h"
#include "wx_dc.h"
#include "wx_dccan.h"
#include "wx_dcmem.h"
#include "wx_dcps.h"
#include "wx_utils.h"
#include "wx_canvs.h"
#include "wx_dialg.h"
#include "wx_main.h"
*/

// Declarations local to this file

#define YSCALE(y) (yorigin - (y))

#define     wx_round(a)    (int)((a)+.5)

// Default constructor
wxDC::wxDC(void)
{
  filename = NULL;
  selected_bitmap = NULL;
  canvas = NULL;
}


wxDC::~wxDC(void)
{
  if (filename)
    delete[] filename;

  if (cdc)
  {
    SelectOldObjects(cdc);
    DeleteDC(cdc);
  }
}

// This will select current objects out of the DC,
// which is what you have to do before deleting the
// DC.
void wxDC::SelectOldObjects(HDC dc)
{
  if (dc)
  {
    if (old_bitmap)
      ::SelectObject(dc, old_bitmap);
    if (old_pen)
      ::SelectObject(dc, old_pen);
    if (old_brush)
      ::SelectObject(dc, old_brush);
    if (old_font)
      ::SelectObject(dc, old_font);
  }
}

void wxDC::SetClippingRegion(float cx, float cy, float cw, float ch)
{
      clipping = TRUE;
      clip_x1 = (int)cx;
      clip_y1 = (int)cy;
      clip_x2 = (int)(cx + cw);
      clip_y2 = (int)(cy + ch);

      HDC dc = NULL;
      wxWnd *wnd = NULL;
      if (canvas) wnd = (wxWnd *)canvas->handle;
      if (cdc)
        dc = cdc;
      else if (wnd)
        dc = GetDC(wnd->handle);

      DoClipping(dc);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::DoClipping(HDC dc)
{
  if (clipping && dc)
  {
    int x_off = 0;
    int y_off = 0;
    if (canvas)
    {
      wxWnd *wnd = (wxWnd *)canvas->handle;

      wnd->CalcScrolledPosition(0, 0, &x_off, &y_off);
    }
//    HRGN rgn = CreateRectRgn(XLOG2DEV(clip_x1 + x_off), YLOG2DEV(clip_y1 + y_off),
//                          XLOG2DEV(clip_x2 + x_off), YLOG2DEV(clip_y2 + y_off));

//    SelectClipRgn(dc, rgn);
//    DeleteObject(rgn);
    IntersectClipRect(dc, XLOG2DEV(clip_x1 + x_off), YLOG2DEV(clip_y1 + y_off),
                          XLOG2DEV(clip_x2 + x_off), YLOG2DEV(clip_y2 + y_off));
  }
}

void wxDC::DestroyClippingRegion(void)
{
      HDC dc = NULL;
      wxWnd *wnd = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;
      if (cdc)
        dc = cdc;
      else if (wnd)
        dc = GetDC(wnd->handle);

      if (clipping && dc)
      {
//        SelectClipRgn(dc, NULL);
        HRGN rgn = CreateRectRgn(0, 0, 32000, 32000);
        SelectClipRgn(dc, rgn);
        DeleteObject(rgn);
      }
      clipping = FALSE;

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::Clear(void)
{
        HDC dc = NULL;
        wxWnd *wnd = NULL;
        if (canvas) wnd = (wxWnd *)canvas->handle;
        if (cdc)
          dc = cdc;
        else
          dc = GetDC(wnd->handle);

        RECT rect;
        if (canvas)
          GetClientRect(wnd->handle, &rect);
        else if (selected_bitmap)
        {
          rect.left = 0; rect.top = 0;
          rect.right = selected_bitmap->GetWidth();
          rect.bottom = selected_bitmap->GetHeight();
        }
        int mode = ::SetMapMode(dc, MM_TEXT);

        DWORD colour = GetBkColor(dc);
        HBRUSH brush = CreateSolidBrush(colour);
        FillRect(dc, &rect, brush);
        DeleteObject(brush);

        ::SetMapMode(dc, MM_ANISOTROPIC);
        SetViewportExtEx(dc, VIEWPORT_EXTENT, VIEWPORT_EXTENT, NULL);
        SetWindowExtEx(dc, window_ext_x, window_ext_y, NULL);

        if (!cdc)
          ReleaseDC(wnd->handle, dc);
}

void wxDC::DrawLine(float x1, float y1, float x2, float y2)
{
      int xx1 = (int)x1;
      int yy1 = (int)y1;
      int xx2 = (int)x2;
      int yy2 = (int)y2;

      if (current_pen)
        SetPen(current_pen);

      wxWnd *wnd = NULL;
      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x1, (int)y1, &xx1, &yy1);
        wnd->CalcScrolledPosition((int)x2, (int)y2, &xx2, &yy2);
      }
      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else if (wnd)
        dc = GetDC(wnd->handle);

      (void)MoveToEx(dc, XLOG2DEV(xx1), YLOG2DEV(yy1), NULL);
      (void)LineTo(dc, XLOG2DEV(xx2), YLOG2DEV(yy2));

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      CalcBoundingBox((float)XLOG2DEV(x1), (float)YLOG2DEV(y1));
      CalcBoundingBox((float)XLOG2DEV(x2), (float)YLOG2DEV(y2));
}

void wxDC::DrawArc(float x1,float y1,float x2,float y2,float xc,float yc)
{
  int xx1 = (int)x1;
  int yy1 = (int)y1;
  int xx2 = (int)x2;
  int yy2 = (int)y2;
  int xxc = (int)xc ;
  int yyc = (int)yc;

  double dx = xc-x1 ;
  double dy = yc-y1 ;
  double radius = (double)sqrt(dx*dx+dy*dy) ;;
  if (xx1==xx2 && xx2==yy2)
  {
    DrawEllipse(xc,yc,(float)(radius*2.0),(float)(radius*2)) ;
    return ;
  }

  if (current_pen)
    SetPen(current_pen) ;

  wxWnd *wnd = NULL;
  if (canvas)
  {
    wnd = (wxWnd *)canvas->handle;
    wnd->CalcScrolledPosition((int)x1, (int)y1, &xx1, &yy1);
    wnd->CalcScrolledPosition((int)x2, (int)y2, &xx2, &yy2);
    wnd->CalcScrolledPosition((int)xc, (int)yc, &xxc, &yyc);
  }
  HDC dc = NULL;
  if (cdc)
    dc = cdc;
  else if (wnd)
    dc = GetDC(wnd->handle);

  xx1 = XLOG2DEV(xx1) ;
  yy1 = YLOG2DEV(yy1) ;
  xx2 = XLOG2DEV(xx2) ;
  yy2 = YLOG2DEV(yy2) ;
  xxc = XLOG2DEV(xxc) ;
  yyc = YLOG2DEV(yyc) ;
  double ray = sqrt((xxc-xx1)*(xxc-xx1)+(yyc-yy1)*(yyc-yy1)) ;
 
  (void)MoveToEx(dc, xx1, yy1, NULL);
  int xxx1 = (int)(xxc-ray);
  int yyy1 = (int)(yyc-ray);
  int xxx2 = (int)(xxc+ray);
  int yyy2 = (int)(yyc+ray);
  if (current_brush && current_brush->style!=wxTRANSPARENT)
  {
    SetBrush(current_brush) ;
    Pie(dc,xxx1,yyy1,xxx2,yyy2,
        xx1,yy1,xx2,yy2) ;
  }
  else
    Arc(dc,xxx1,yyy1,xxx2,yyy2,
        xx1,yy1,xx2,yy2) ;

  if (!cdc)
    ReleaseDC(wnd->handle, dc);

  CalcBoundingBox((float)XLOG2DEV(xc-radius), (float)YLOG2DEV(yc-radius));
  CalcBoundingBox((float)XLOG2DEV(xc+radius), (float)YLOG2DEV(yc+radius));
}

void wxDC::DrawPoint(float x, float y)
{
      wxWnd *wnd = NULL;

      if (current_pen)
        SetPen(current_pen) ;

      int xx1 = (int)x;
      int yy1 = (int)y;
      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &xx1, &yy1);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

/*    (void)MoveToEx(dc, XLOG2DEV(xx1), YLOG2DEV(yy1), NULL);
      (void)LineTo(dc, XLOG2DEV(xx1), YLOG2DEV(yy1));
*/
      COLORREF color = 0x00ffffff;
      if (current_pen)
        color = RGB(current_pen->GetColour().Red(),
                    current_pen->GetColour().Green(),
                    current_pen->GetColour().Blue());

      SetPixel(dc, XLOG2DEV(xx1), YLOG2DEV(yy1), color);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      CalcBoundingBox((float)XLOG2DEV(x), (float)YLOG2DEV(y));
}

void wxDC::DrawPolygon(int n, wxPoint points[], float xoffset, float yoffset)
{
      int xoffset1 = 0;
      int yoffset1 = 0;
      wxWnd *wnd = NULL;

      if (current_pen)
        SetPen(current_pen) ;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition(0, 0, &xoffset1, &yoffset1);
      }

      xoffset1 += (int)xoffset; yoffset1 += (int)yoffset;

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      POINT *cpoints = new POINT[n];
      int i;
      for (i = 0; i < n; i++)
      {
        cpoints[i].x = (int)(XLOG2DEV(points[i].x + xoffset1));
        cpoints[i].y = (int)(YLOG2DEV(points[i].y + yoffset1));

        CalcBoundingBox((float)cpoints[i].x, (float)cpoints[i].y);
      }

      (void)Polygon(dc, cpoints, n);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      delete[] cpoints;
}

void wxDC::DrawLines(int n, wxPoint points[], float xoffset, float yoffset)
{
      int xoffset1 = 0;
      int yoffset1 = 0;
      wxWnd *wnd = NULL;

      if (current_pen)
        SetPen(current_pen) ;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition(0, 0, &xoffset1, &yoffset1);
      }
      xoffset1 += (int)xoffset; yoffset1 += (int)yoffset;

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      POINT *cpoints = new POINT[n];
      int i;
      for (i = 0; i < n; i++)
      {
        cpoints[i].x = (int)(XLOG2DEV(points[i].x + xoffset1));
        cpoints[i].y = (int)(YLOG2DEV(points[i].y + yoffset1));

        CalcBoundingBox((float)cpoints[i].x, (float)cpoints[i].y);
      }

      (void)Polyline(dc, cpoints, n);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      delete[] cpoints;
}

void wxDC::DrawRectangle(float x, float y, float width, float height)
{
      wxWnd *wnd = NULL;

      int x1 = (int)x;
      int y1 = (int)y;
      int x2 = (int)(x+width);
      int y2 = (int)(y+height);

      if (current_pen)
        SetPen(current_pen) ;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &x1, &y1);
        wnd->CalcScrolledPosition((int)(x+width), (int)(y+height), &x2, &y2);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      (void)Rectangle(dc, XLOG2DEV(x1), YLOG2DEV(y1), XLOG2DEV(x2), YLOG2DEV(y2));

      CalcBoundingBox((float)XLOG2DEV(x1), (float)YLOG2DEV(y1));
      CalcBoundingBox((float)XLOG2DEV(x2), (float)YLOG2DEV(y2));

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::DrawRoundedRectangle(float x, float y, float width, float height, float radius)
{
      wxWnd *wnd = NULL;

      int x1 = (int)x;
      int y1 = (int)y;
      int x2 = (int)(x+width);
      int y2 = (int)(y+height);

      if (current_pen)
        SetPen(current_pen) ;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &x1, &y1);
        wnd->CalcScrolledPosition((int)(x+width), (int)(y+height), &x2, &y2);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      (void)RoundRect(dc, XLOG2DEV(x1), YLOG2DEV(y1), XLOG2DEV(x2),
                          YLOG2DEV(y2), XLOG2DEV(radius), YLOG2DEV(radius));

      CalcBoundingBox((float)XLOG2DEV(x1), (float)YLOG2DEV(y1));
      CalcBoundingBox((float)XLOG2DEV(x2), (float)YLOG2DEV(y2));

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::DrawEllipse(float x, float y, float width, float height)
{
      wxWnd *wnd = NULL;

      int x1 = (int)x;
      int y1 = (int)y;
      int x2 = (int)(x+width);
      int y2 = (int)(y+height);

      if (current_pen)
        SetPen(current_pen) ;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &x1, &y1);
        wnd->CalcScrolledPosition((int)(x+width), (int)(y+height), &x2, &y2);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      (void)Ellipse(dc, XLOG2DEV(x1), YLOG2DEV(y1), XLOG2DEV(x2), YLOG2DEV(y2));

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      CalcBoundingBox((float)XLOG2DEV(x1), (float)YLOG2DEV(y1));
      CalcBoundingBox((float)XLOG2DEV(x2), (float)YLOG2DEV(y2));
}

void wxDC::DrawIcon(wxIcon *icon, float x, float y)
{
      wxWnd *wnd = NULL;

      int x1 = (int)x;
      int y1 = (int)y;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &x1, &y1);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      ::DrawIcon(dc, XLOG2DEV(x1), YLOG2DEV(y1), icon->ms_icon);
      CalcBoundingBox((float)XLOG2DEV(x1), (float)YLOG2DEV(y1));

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}


void wxDC::SetFont(wxFont *the_font)
{
  font = the_font;

      wxWnd *wnd = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      if (font)
        font->BuildInternalFont(dc) ;

      if (font && font->cfont)
      {
        HFONT f = ::SelectObject(dc, font->cfont);
        if (!old_font)
          old_font = f;
      }
      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::SetPen(wxPen *pen)
{
  current_pen = pen;
  if (!pen)
  {
    return;
  }
      HDC dc = NULL;
      wxWnd *wnd = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;
      if (cdc)
        dc = cdc;
      else if (wnd)
        dc = GetDC(wnd->handle);

      HPEN p = ::SelectObject(dc, pen->cpen);
      if (!old_pen)
        old_pen = p;

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::SetBrush(wxBrush *brush)
{
  current_brush = brush;
  if (!brush)
  {
    return;
  }
      HDC dc = NULL;
      wxWnd *wnd = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      if (current_brush)
      {
        HBRUSH b;
        if (current_brush->cbrush)
          b = ::SelectObject(dc, current_brush->cbrush);
        else
          b = ::SelectObject(dc, ::GetStockObject(NULL_BRUSH));
        if (!old_brush)
          old_brush = b;
      }
      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::DrawText(char *text, float x, float y)
{
      wxWnd *wnd = NULL;

      int xx1 = (int)x;
      int yy1 = (int)y;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
        wnd->CalcScrolledPosition((int)x, (int)y, &xx1, &yy1);
      }

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      if (font && font->cfont)
      {
        HFONT f = ::SelectObject(dc, font->cfont);
        if (!old_font)
          old_font = f;
      }

      if (current_text_foreground.Ok())
        SetTextColor(dc, RGB(current_text_foreground.Red(),
                             current_text_foreground.Green(),
                             current_text_foreground.Blue()));

      DWORD old_background;
      if (current_text_background.Ok())
      {
        old_background = SetBkColor(dc, RGB(current_text_background.Red(),
                                            current_text_background.Green(),
                                            current_text_background.Blue()));
      }
      if (current_bk_mode == wxTRANSPARENT)
        SetBkMode(dc, TRANSPARENT);
      else
        SetBkMode(dc, OPAQUE);

      (void)TextOut(dc, XLOG2DEV(xx1), YLOG2DEV(yy1), text, strlen(text));

      if (current_text_background.Ok())
        (void)SetBkColor(dc, old_background);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      CalcBoundingBox((float)XLOG2DEV(xx1), (float)YLOG2DEV(yy1));
}

void wxDC::SetBackground(wxBrush *brush)
{
  current_background_brush = brush;
  if (canvas)
  {
    wxCanvasWnd *wnd = (wxCanvasWnd *)canvas->handle;
    wnd->background_brush = brush->cbrush;
  }
  wxWnd *wnd = NULL;

  if (canvas)
    wnd = (wxWnd *)canvas->handle;

  HDC dc = NULL;
  if (cdc)
    dc = cdc;
  else
    dc = GetDC(wnd->handle);

  if (current_background_brush)
    (void)SetBkColor(dc, RGB(current_background_brush->colour->Red(),
                             current_background_brush->colour->Green(),
                             current_background_brush->colour->Blue()));

  if (!cdc)
    ReleaseDC(wnd->handle, dc);
}

void wxDC::SetLogicalFunction(int function)
{
      current_logical_function = function;

      HDC dc = NULL;
      wxWnd *wnd = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      SetRop(dc);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
}

void wxDC::SetRop(HDC dc)
{
  if (!dc || current_logical_function < 0)
    return;

  int c_rop;
  // These may be wrong
  switch (current_logical_function)
  {
//    case wxXOR: c_rop = R2_XORPEN; break;
    case wxXOR: c_rop = R2_NOTXORPEN; break;
    case wxINVERT: c_rop = R2_NOT; break;
    case wxOR_REVERSE: c_rop = R2_MERGEPENNOT; break;
    case wxAND_REVERSE: c_rop = R2_MASKPENNOT; break;
    case wxCLEAR: c_rop = R2_WHITE; break;
    case wxSET: c_rop = R2_BLACK; break;
    case wxSRC_INVERT: c_rop = R2_NOTCOPYPEN; break;
    case wxOR_INVERT: c_rop = R2_MERGENOTPEN; break;
    case wxAND: c_rop = R2_MASKPEN; break;
    case wxOR: c_rop = R2_MERGEPEN; break;
    case wxAND_INVERT: c_rop = R2_MASKNOTPEN; break;
    case wxEQUIV:
    case wxNAND:
    case wxCOPY:
    default:
      c_rop = R2_COPYPEN; break;
  }
  SetROP2(dc, c_rop);
}

Bool wxDC::StartDoc(char *message)
{
  Bool flag = FALSE;

  DOCINFO docinfo;
  docinfo.cbSize = sizeof(DOCINFO);
  docinfo.lpszDocName = message;
  docinfo.lpszOutput = filename;
  if (cdc) flag = ::StartDoc(cdc, &docinfo);
  else flag = FALSE;

  return flag;
}

void wxDC::EndDoc(void)
{
  if (cdc) ::EndDoc(cdc);
}

void wxDC::StartPage(void)
{
  if (cdc)
    ::StartPage(cdc);
}

void wxDC::EndPage(void)
{
  if (cdc)
    ::EndPage(cdc);
}

float wxDC::GetCharHeight(void)
{
      TEXTMETRIC lpTextMetric;
      wxWnd *wnd = NULL;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
      }

      HDC dc = NULL;

      if (cdc)
        dc = cdc;
      else dc = GetDC(wnd->handle);

      GetTextMetrics(dc, &lpTextMetric);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      return (float)YDEV2LOGREL(lpTextMetric.tmHeight);
}

float wxDC::GetCharWidth(void)
{
      TEXTMETRIC lpTextMetric;
      wxWnd *wnd = NULL;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
      }

      HDC dc = NULL;

      if (cdc)
        dc = cdc;
      else dc = GetDC(wnd->handle);

      GetTextMetrics(dc, &lpTextMetric);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      return (float)XDEV2LOGREL(lpTextMetric.tmAveCharWidth);
}

void wxDC::GetTextExtent(char *string, float *x, float *y,
                         float *descent, float *externalLeading)
{
      wxWnd *wnd = NULL;

      if (canvas)
      {
        wnd = (wxWnd *)canvas->handle;
      }

      HDC dc = NULL;

      if (cdc)
        dc = cdc;
      else dc = GetDC(wnd->handle);

      SIZE sizeRect;
      TEXTMETRIC tm;
      GetTextExtentPoint(dc, string, strlen(string), &sizeRect);
      GetTextMetrics(dc, &tm);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);

      *x = (float)XDEV2LOGREL(sizeRect.cx);
      *y = (float)YDEV2LOGREL(sizeRect.cy);
      if (descent) *descent = (float)tm.tmDescent;
      if (externalLeading) *externalLeading = (float)tm.tmExternalLeading;
}

void wxDC::SetMapMode(int mode)
{
  mapping_mode = mode;

  int pixel_width = 0;
  int pixel_height = 0;
  int mm_width = 0;
  int mm_height = 0;

  HDC dc = NULL;
  wxWnd *wnd = NULL;
  if (canvas)
    wnd = (wxWnd *)canvas->handle;

  if (cdc)
    dc = cdc;
  else if (wnd)
    dc = GetDC(wnd->handle);

  pixel_width = GetDeviceCaps(dc, HORZRES);
  pixel_height = GetDeviceCaps(dc, VERTRES);
  mm_width = GetDeviceCaps(dc, HORZSIZE);
  mm_height = GetDeviceCaps(dc, VERTSIZE);

  if ((pixel_width == 0) || (pixel_height == 0) || (mm_width == 0) || (mm_height == 0))
  {
    if (!cdc)
      ReleaseDC(wnd->handle, dc);
    return;
  }

  float mm2pixelsX = (float)pixel_width/mm_width;
  float mm2pixelsY = (float)pixel_height/mm_height;

  switch (mode)
  {
    case MM_TWIPS:
    {
      logical_scale_x = (float)(twips2mm * mm2pixelsX);
      logical_scale_y = (float)(twips2mm * mm2pixelsY);
      break;
    }
    case MM_POINTS:
    {
      logical_scale_x = (float)(pt2mm * mm2pixelsX);
      logical_scale_y = (float)(pt2mm * mm2pixelsY);
      break;
    }
    case MM_METRIC:
    {
      logical_scale_x = mm2pixelsX;
      logical_scale_y = mm2pixelsY;
      break;
    }
    case MM_LOMETRIC:
    {
      logical_scale_x = (float)(mm2pixelsX/10.0);
      logical_scale_y = (float)(mm2pixelsY/10.0);
      break;
    }
    default:
    case MM_TEXT:
    {
      logical_scale_x = 1.0;
      logical_scale_y = 1.0;
      break;
    }
  }

  ::SetMapMode(dc, MM_ANISOTROPIC);
  SetViewportExtEx(dc, VIEWPORT_EXTENT, VIEWPORT_EXTENT, NULL);
  window_ext_x = MS_XDEV2LOG(VIEWPORT_EXTENT);
  window_ext_y = MS_YDEV2LOG(VIEWPORT_EXTENT);
  SetWindowExtEx(dc, window_ext_x, window_ext_y, NULL);

  if (!cdc)
    ReleaseDC(wnd->handle, dc);
}

void wxDC::SetUserScale(float x, float y)
{
  user_scale_x = x;
  user_scale_y = y;

  SetMapMode(mapping_mode);
}

float wxDC::DeviceToLogicalX(int x)
{
  return (float)MS_XDEV2LOG(x);
}

float wxDC::DeviceToLogicalXRel(int x)
{
  return (float)MS_XDEV2LOGREL(x);
}

float wxDC::DeviceToLogicalY(int y)
{
  return (float)MS_YDEV2LOG(y);
}

float wxDC::DeviceToLogicalYRel(int y)
{
  return (float)MS_YDEV2LOGREL(y);
}

int wxDC::LogicalToDeviceX(float x)
{
  return MS_XLOG2DEV(x);
}

int wxDC::LogicalToDeviceXRel(float x)
{
  return MS_XLOG2DEVREL(x);
}

int wxDC::LogicalToDeviceY(float y)
{
  return MS_YLOG2DEV(y);
}

int wxDC::LogicalToDeviceYRel(float y)
{
  return MS_YLOG2DEVREL(y);
}

Bool wxDC::Blit(float xdest, float ydest, float width, float height,
                wxCanvasDC *source, float xsrc, float ysrc, int rop)
{
      wxWnd *wnd = NULL;
      wxWnd *wnd_src = NULL;
      if (canvas)
        wnd = (wxWnd *)canvas->handle;
      if (source->canvas)
        wnd_src = (wxWnd *)source->canvas->handle;

      HDC dc = NULL;
      if (cdc)
        dc = cdc;
      else
        dc = GetDC(wnd->handle);

      HDC dc_src = NULL;
      if (source->cdc)
        dc_src = source->cdc;
      else if (wnd)
        dc_src = GetDC(wnd_src->handle);

      int xdest1 = (int)xdest;
      int ydest1 = (int)ydest;
      int xsrc1 = (int)xsrc;
      int ysrc1 = (int)ysrc;

      if (canvas)
      {
        wnd->CalcScrolledPosition((int)xdest, (int)ydest, &xdest1, &ydest1);
      }
      if (source->canvas)
      {
        wnd_src->CalcScrolledPosition((int)xsrc, (int)ysrc, &xsrc1, &ysrc1);
      }

      Bool success = BitBlt(dc, xdest1, ydest1, (int)width, (int)height, dc_src,
                                xsrc1, ysrc1,

                rop == wxCOPY ? SRCCOPY :
                rop == wxCLEAR ? WHITENESS :
                rop == wxSET ? BLACKNESS :
                rop == wxINVERT ? DSTINVERT :
                rop == wxAND ? MERGECOPY :
                rop == wxOR ? MERGEPAINT :
                rop == wxSRC_INVERT ? NOTSRCCOPY :
                rop == wxXOR ? SRCINVERT :
                rop == wxOR_REVERSE ? MERGEPAINT :
                rop == wxAND_REVERSE ? SRCERASE :
                SRCCOPY);

      if (!cdc)
        ReleaseDC(wnd->handle, dc);
      if (!source->cdc)
        ReleaseDC(wnd_src->handle, dc_src);
      return success;
}

wxCanvasDC::wxCanvasDC(void)
{
  canvas = NULL;
  device = wxDEVICE_WINDOWS;
  font = NULL;
  logical_origin_x = 0;
  logical_origin_y = 0;
  device_origin_x = 0;
  device_origin_y = 0;
  logical_scale_x = 1.0;
  logical_scale_y = 1.0;
  user_scale_x = 1.0;
  user_scale_y = 1.0;
  mapping_mode = MM_TEXT;
  title = NULL;
  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;
  dont_delete = FALSE;
  cdc = NULL;
  clipping = FALSE;
  ok = TRUE;
  window_ext_x = VIEWPORT_EXTENT;
  window_ext_y = VIEWPORT_EXTENT;
  current_logical_function = -1;
  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;
  Colour = wxColourDisplay();
}

wxCanvasDC::wxCanvasDC(wxCanvas *the_canvas):wxbCanvasDC(the_canvas)
{
  canvas = the_canvas;
  device = wxDEVICE_WINDOWS;
  min_x = 0; min_y = 0; max_x = 0; max_y = 0;
  font = NULL;

  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  title = NULL;

  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;
  dont_delete = FALSE;
  cdc = NULL;
  clipping = FALSE;
  ok = TRUE;
  window_ext_x = VIEWPORT_EXTENT;
  window_ext_y = VIEWPORT_EXTENT;

  wxWnd *wnd = (wxWnd *)the_canvas->handle;
  HDC dc = GetDC(wnd->handle);
  SetMapMode(MM_TEXT);

  ReleaseDC(wnd->handle, dc);

  current_logical_function = -1;

  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;
  Colour = wxColourDisplay();
  SetBrush(wxWHITE_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxCanvasDC::~wxCanvasDC(void)
{
}

void wxCanvasDC::GetClippingBox(float *x,float *y,float *w,float *h)
{
  if (clipping)
  {
    *x = (float)clip_x1 ;
    *y = (float)clip_y1 ;
    *w = (float)(clip_x2 - clip_x1) ;
    *h = (float)(clip_y2 - clip_y1) ;
  }
  else
   *x = *y = *w = *h = 0 ;

}

wxPrinterDC::wxPrinterDC(char *driver_name, char *device_name, char *file, Bool interactive)
{
  min_x = 0; min_y = 0; max_x = 0; max_y = 0;
  wx_interactive = interactive;
  font = NULL;
  device = wxDEVICE_WINDOWS;
  canvas = NULL;

  cdc = NULL;
  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;
  dont_delete = FALSE;
  clipping = FALSE;

  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  min_x = 1000.0;
  min_y = 1000.0;
  max_x = -1000.0;
  max_y = -1000.0;
  title = NULL;
  if (file)
    filename = copystring(file);
  else filename = NULL;

   if (!driver_name || !device_name || !file)
   {
     PRINTDLG pd;
	
     pd.lStructSize = sizeof( PRINTDLG );
     pd.hwndOwner=NULL;
     pd.hDevMode=(HANDLE)NULL;
     pd.hDevNames=(HANDLE)NULL;
     pd.Flags=PD_RETURNDC | PD_NOSELECTION | PD_NOPAGENUMS;
     pd.nFromPage=0;
     pd.nToPage=0;
     pd.nMinPage=0;
     pd.nMaxPage=0;
     pd.nCopies=1;
     pd.hInstance=(HINSTANCE)NULL;

     if ( PrintDlg( &pd ) != 0 )
     {
       cdc = pd.hDC;
       ok = TRUE;
     }
     else
     {
       ok = FALSE;
       return;
     }

     dont_delete = TRUE; // ??? WHY???
   }
   else
   {
     cdc = CreateDC(driver_name, device_name, file, NULL);
     ok = TRUE;
   }
   if (cdc)
   {
//     int width = GetDeviceCaps(cdc, VERTRES);
//     int height = GetDeviceCaps(cdc, HORZRES);
     SetMapMode(MM_TEXT);
   }

  current_logical_function = -1;
  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;
  Colour = wxColourDisplay();

  SetBrush(wxBLACK_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxPrinterDC::~wxPrinterDC(void)
{
}

/*
 * Memory DC
 *
 */

wxMemoryDC::wxMemoryDC(void)
{
  filename = NULL;
  clipping = FALSE;
  dont_delete = FALSE;
  window_ext_x = VIEWPORT_EXTENT;
  window_ext_y = VIEWPORT_EXTENT;
  device = wxDEVICE_WINDOWS;
  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;
  canvas = NULL;

  ok = FALSE;
  title = NULL;

  current_logical_function = -1;
  font = NULL;
  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;

  cdc = ::CreateCompatibleDC(NULL);
  ok = (cdc != NULL);

  Colour = wxColourDisplay();
  SetBrush(wxWHITE_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxMemoryDC::wxMemoryDC(wxCanvasDC *old_dc):wxbMemoryDC(old_dc)
{
  min_x = 0; min_y = 0; max_x = 0; max_y = 0;
  filename = NULL;
  clipping = FALSE;
  dont_delete = FALSE;
  window_ext_x = VIEWPORT_EXTENT;
  window_ext_y = VIEWPORT_EXTENT;
  device = wxDEVICE_WINDOWS;
  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;
  canvas = NULL;

  ok = FALSE;
  title = NULL;

  current_logical_function = -1;
  font = NULL;
  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  current_pen = NULL;
  current_brush = NULL;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
//  current_text_background = NULL;

  wxWnd *wnd = NULL;
  if (old_dc->canvas)
    wnd = (wxWnd *)old_dc->canvas->handle;

  HDC dc = NULL;
  if (old_dc->cdc)
    dc = old_dc->cdc;
  else
    dc = GetDC(wnd->handle);

  cdc = ::CreateCompatibleDC(dc);
  ok = (cdc != NULL);

  if (!old_dc->cdc)
    ReleaseDC(wnd->handle, dc);

  Colour = wxColourDisplay();
  SetBrush(wxWHITE_BRUSH);
  SetPen(wxBLACK_PEN);
}

wxMemoryDC::~wxMemoryDC(void)
{
}

void wxMemoryDC::SelectObject(wxBitmap *bitmap)
{
  selected_bitmap = bitmap;
  HBITMAP bm = ::SelectObject(cdc, bitmap->ms_bitmap);
  if (bm == ERROR)
  {
    wxMessageBox("Error in wxMemoryDC::SelectObject\nBitmap may be selected in another wxMemoryDC.\nDelete the first wxMemoryDC to deselect bitmap.",
                 "wxWindows Error");
  }
  else if (!old_bitmap)
    old_bitmap = bm;
}

