/*
 * File:     wx_gdi.cc
 * Purpose:  GDI (Graphics Device Interface) objects and functions
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "wx.h"
#pragma hdrstop
#include <iostream.h>

/*
#include <windows.h>
#include "common.h"
#include "wx_list.h"
#include "wx_utils.h"
#include "wx_main.h"
#include "wx_gdi.h"
*/

wxFont::wxFont(void)
{
  point_size = 0;
  temporary = TRUE;
  cfont = NULL;
}

/* Constructor for a font. Note that the real construction is done
 * in wxDC::SetFont, when information is available about scaling etc.
 */
wxFont::wxFont(int PointSize, int Family, int Style, int Weight, Bool Underlined):
  wxbFont(PointSize, Family, Style, Weight, Underlined)
{
  family = Family;
  style = Style;
  weight = Weight;
  point_size = PointSize;
  underlined = Underlined;

  temporary = FALSE;
  cfont = NULL;

  wxTheFontList->Append(this);
}

wxFont::~wxFont()
{
  if (cfont)
    DeleteObject(cfont);
  wxTheFontList->DeleteObject(this);
}

void wxFont::BuildInternalFont(HDC dc)
{

  if (!cfont)
  {
    BYTE ff_italic;
    int ff_weight = 0;
    int ff_family = 0;
    char *ff_face = NULL;

    switch (family)
    {
      case wxDECORATIVE: ff_family = FF_DECORATIVE;
                         break;
      case wxROMAN:      ff_family = FF_ROMAN;
                         break;
      case wxMODERN:     ff_family = FF_MODERN;
                         break;
      case wxSWISS:      ff_family = FF_SWISS;
                         ff_face = "Arial";
                         break;
      case wxDEFAULT:
      default:           ff_family = FF_DONTCARE;
    }

    if (style == wxITALIC || style == wxSLANT)
      ff_italic = 1;
    else
      ff_italic = 0;

    if (weight == wxNORMAL)
      ff_weight = FW_NORMAL;
    else if (weight == wxLIGHT)
      ff_weight = FW_LIGHT;
    else if (weight == wxBOLD)
      ff_weight = FW_BOLD;

    Bool ff_underline = underlined;
    int nHeight = point_size*GetDeviceCaps(dc, LOGPIXELSY)/72;
    cfont = CreateFont(nHeight, 0, 0, 0,ff_weight,ff_italic,(BYTE)ff_underline,
                0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                PROOF_QUALITY, DEFAULT_PITCH | ff_family, ff_face);
  }

}

HFONT wxFont::GetInternalFont(HDC dc)
{
  BuildInternalFont(dc) ;
  return cfont ;
}

/*
 * Colour map
 *
 */

wxColourMap::wxColourMap(void)
{
}

wxColourMap::~wxColourMap(void)
{
}

// Pens

wxPen::wxPen(void)
{
  colour = NULL;
  style = wxSOLID;
  width = 1;
  cpen = NULL;
  wxThePenList->AddPen(this);
}

wxPen::~wxPen()
{
  if (colour)
    delete colour;

  if (cpen)
    DeleteObject(cpen);

  wxThePenList->RemovePen(this);
}

wxPen::wxPen(wxColour& col, int Width, int Style)
{
  colour = new wxColour(col.Red(), col.Green(), col.Blue());
  width = Width;
  style = Style;

  DWORD ms_colour = RGB(col.Red(), col.Green(), col.Blue());
  cpen = CreatePen(wx2msPenStyle(Style), Width, ms_colour);
  wxThePenList->AddPen(this);
}

wxPen::wxPen(char *col, int Width, int Style)
{
  colour = new wxColour(col);
  width = Width;
  style = Style;

  DWORD ms_colour = RGB(colour->Red(), colour->Green(), colour->Blue());
  cpen = CreatePen(wx2msPenStyle(Style), Width, ms_colour);

  wxThePenList->AddPen(this);
}

void wxPen::SetColour(wxColour& col)
{
  // Must allocate BEFORE deleting, since &col can be a ref to this->colour!
  wxColour *newcol = new wxColour(col.Red(),col.Green(),col.Blue()) ;
  if (colour)
    delete colour;
  colour = newcol ;

  if (cpen)
    DeleteObject(cpen);

  DWORD ms_colour = RGB(col.Red(), col.Green(), col.Blue());
  cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
}

void wxPen::SetColour(char *col)
{
  if (colour)
    delete colour;
  colour = new wxColour(col);

  if (cpen)
    DeleteObject(cpen);

  DWORD ms_colour = RGB(colour->Red(), colour->Green(), colour->Blue());
  cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
}

void wxPen::SetColour(char red, char green, char blue)
{
  if (colour)
    delete colour;
  colour = new wxColour(red, green, blue);

  if (cpen)
    DeleteObject(cpen);

  DWORD ms_colour = RGB(red, green, blue);
  cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
}

void wxPen::SetWidth(int Width)
{
  width = Width;

  if (cpen)
    DeleteObject(cpen);

  DWORD ms_colour = 0;
  if (colour)
    ms_colour = RGB(colour->Red(), colour->Green(), colour->Blue());
  cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
}

void wxPen::SetStyle(int Style)
{
  style = Style;

  if (cpen)
    DeleteObject(cpen);

  DWORD ms_colour = 0;
  if (colour)
    ms_colour = RGB(colour->Red(), colour->Green(), colour->Blue());
  cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
}

int wx2msPenStyle(int wx_style)
{
  int cstyle;
  switch (wx_style)
  {
    case wxDOT:
      cstyle = PS_DOT;
      break;
    case wxSHORT_DASH:
    case wxLONG_DASH:
      cstyle = PS_DASH;
      break;
    case wxTRANSPARENT:
      cstyle = PS_NULL;
      break;
    case wxSOLID:
    default:
      cstyle = PS_SOLID;
      break;
  }
  return cstyle;
}


// Brushes
wxBrush::wxBrush(void)
{
  colour = NULL;
  style = wxSOLID;
  cbrush = NULL;

  wxTheBrushList->AddBrush(this);
}

wxBrush::~wxBrush()
{
  if (colour)
    delete colour;

  if (cbrush)
    DeleteObject(cbrush);

  wxTheBrushList->RemoveBrush(this);
}

wxBrush::wxBrush(wxColour& col, int Style)
{
  colour = new wxColour(col.Red(), col.Green(), col. Blue());
  style = Style;

  switch (Style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;  // Must always select a suitable background brush
                      // - could choose white always for a quick solution
      break;
    case wxSOLID:
    default:
      cbrush = CreateSolidBrush(RGB(col.Red(), col.Green(), col.Blue()));
  }

  wxTheBrushList->AddBrush(this);
}

wxBrush::wxBrush(char *col, int Style)
{
  colour = new wxColour(col);
  style = Style;

  switch (Style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;  // Must always select a suitable background brush
                      // - could choose white always for a quick solution
      break;
    case wxSOLID:
    default:
      cbrush = CreateSolidBrush(RGB(colour->Red(), colour->Green(), colour->Blue()));
  }

  wxTheBrushList->AddBrush(this);
}

void wxBrush::SetColour(wxColour& col)
{
  // Must allocate BEFORE deleting, since &col can be a ref to this->colour!
  wxColour *newcol = new wxColour(col.Red(),col.Green(),col.Blue()) ;
  if (colour)
    delete colour;
  colour = newcol ;

  if (cbrush) DeleteObject(cbrush);
  switch (style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;  // Must always select a suitable background brush
                      // - could choose white always for a quick solution
      break;
    case wxSOLID:
    default:
      cbrush = CreateSolidBrush(RGB(col.Red(), col.Green(), col.Blue()));
  }
}

void wxBrush::SetColour(char *col)
{
  if (colour)
    delete colour;
  colour = new wxColour(col);

  if (cbrush) DeleteObject(cbrush);
  switch (style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;  // Must always select a suitable background brush
                      // - could choose white always for a quick solution
      break;
    case wxSOLID:
    default:
      cbrush = CreateSolidBrush(RGB(colour->Red(), colour->Green(), colour->Blue()));
  }
}

void wxBrush::SetColour(char red, char green, char blue)
{
  if (colour)
    delete colour;
  colour = new wxColour(red, green, blue);

  if (cbrush) DeleteObject(cbrush);
  switch (style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;  // Must always select a suitable background brush
                      // - could choose white always for a quick solution
      break;
    case wxSOLID:
    default:
      cbrush = CreateSolidBrush(RGB(red, green, blue));
  }
}

void wxBrush::SetStyle(int Style)
{
  style = Style;

  if (cbrush) DeleteObject(cbrush);
  switch (Style)
  {
    case wxTRANSPARENT:
      cbrush = NULL;     // Don't fill shapes
      break;
    case wxSOLID:
    default:
      if (colour)
      {
        cbrush = CreateSolidBrush(RGB(colour->Red(), colour->Green(), colour->Blue()));
      }
      else cbrush = NULL;
  }
}

// Icons
wxIcon::wxIcon(char bits[], int width, int height):
  wxbIcon(bits, width, height)
{
}

wxIcon::wxIcon(char *icon_file):
  wxbIcon(icon_file)
{
  ms_icon = LoadIcon(wxhInstance, icon_file);

  wxTheIconList->Append(this);
}

wxIcon::~wxIcon(void)
{
  if (ms_icon)
    DestroyIcon(ms_icon);

  wxTheIconList->DeleteObject(this);
}

// Cursors
// NB Should delete loaded Windows cursors with DestroyCursor!!!
wxCursor::wxCursor(char bits[], int width, int height):
  wxbCursor(bits, width, height)
{
}

wxCursor::wxCursor(char *cursor_file):
  wxbCursor(cursor_file)
{
  ms_cursor = LoadCursor(wxhInstance, cursor_file);
}

// Cursors by stock number
wxCursor::wxCursor(int cursor_type):
  wxbCursor(cursor_type)
{
  switch (cursor_type)
  {
    case wxCURSOR_WAIT:
      ms_cursor = LoadCursor(NULL, IDC_WAIT);
      break;
    case wxCURSOR_IBEAM:
      ms_cursor = LoadCursor(NULL, IDC_IBEAM);
      break;
    case wxCURSOR_CROSS:
      ms_cursor = LoadCursor(NULL, IDC_CROSS);
      break;
    case wxCURSOR_SIZENWSE:
      ms_cursor = LoadCursor(NULL, IDC_SIZENWSE);
      break;
    case wxCURSOR_SIZENESW:
      ms_cursor = LoadCursor(NULL, IDC_SIZENESW);
      break;
    case wxCURSOR_SIZEWE:
      ms_cursor = LoadCursor(NULL, IDC_SIZEWE);
      break;
    case wxCURSOR_SIZENS:
      ms_cursor = LoadCursor(NULL, IDC_SIZENS);
      break;
    case wxCURSOR_CHAR:
    {
      ms_cursor = LoadCursor(NULL, IDC_ARROW);
      break;
    }
    case wxCURSOR_HAND:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_HAND");
      break;
    }
    case wxCURSOR_BULLSEYE:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_BULLSEYE");
      break;
    }
    case wxCURSOR_PENCIL:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_PENCIL");
      break;
    }
    case wxCURSOR_MAGNIFIER:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_MAGNIFIER");
      break;
    }
    case wxCURSOR_NO_ENTRY:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_NO_ENTRY");
      break;
    }
    case wxCURSOR_LEFT_BUTTON:
    {
      ms_cursor = LoadCursor(NULL, IDC_ARROW);
      break;
    }
    case wxCURSOR_RIGHT_BUTTON:
    {
      ms_cursor = LoadCursor(NULL, IDC_ARROW);
      break;
    }
    case wxCURSOR_MIDDLE_BUTTON:
    {
      ms_cursor = LoadCursor(NULL, IDC_ARROW);
      break;
    }
    case wxCURSOR_SIZING:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_SIZING");
      break;
    }
    case wxCURSOR_WATCH:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_WATCH");
      break;
    }
    case wxCURSOR_SPRAYCAN:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_ROLLER");
      break;
    }
    case wxCURSOR_PAINT_BRUSH:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_PBRUSH");
      break;
    }
    case wxCURSOR_POINT_LEFT:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_PLEFT");
      break;
    }
    case wxCURSOR_POINT_RIGHT:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_PRIGHT");
      break;
    }
    case wxCURSOR_QUESTION_ARROW:
    {
      ms_cursor = LoadCursor(wxhInstance, "wxCURSOR_QARROW");
      break;
    }
    default:
    case wxCURSOR_ARROW:
      ms_cursor = LoadCursor(NULL, IDC_ARROW);
      break;
  }
}

wxCursor::~wxCursor(void)
{
}

// Global cursor setting
void wxSetCursor(wxCursor *cursor)
{
  if (cursor)
    ::SetCursor(cursor->ms_cursor);

  wxFlushEvents();
}

// Misc. functions

// Return TRUE if we have a colour display
Bool wxColourDisplay(void)
{
  HDC dc = ::GetDC(NULL);
  Bool flag;
  if (GetDeviceCaps(dc, NUMCOLORS) > 2)
    flag = TRUE;
  else
    flag = FALSE;
  ReleaseDC(NULL, dc);
  return flag;
}

// Returns depth of screen
int wxDisplayDepth(void)
{
  HDC dc = ::GetDC(NULL);
  int depth = GetDeviceCaps(dc, PLANES);
  ReleaseDC(NULL, dc);
  return depth;
}

// Get size of display
void wxDisplaySize(int *width, int *height)
{
  HDC dc = ::GetDC(NULL);
  *width = GetDeviceCaps(dc, HORZRES); *height = GetDeviceCaps(dc, VERTRES);
  ReleaseDC(NULL, dc);
}

wxBitmap::wxBitmap(void):
  wxbBitmap(0, 0, 0)
{
  ok = FALSE;
  width = 0;
  height = 0;
  depth = 0;
}

wxBitmap::wxBitmap(char bits[], int the_width, int the_height, int no_bits):
  wxbBitmap(bits, the_width, the_height, no_bits)
{
  ok = FALSE;
}

wxBitmap::wxBitmap(int w, int h, int d):
  wxbBitmap(w, h, d)
{
  width = w;
  height = h;
  depth = d;

  if (d > 0)
  {
    ms_bitmap = CreateBitmap(w, h, d, 1, NULL);
  }
  else
  {
    HDC dc = GetDC(NULL);
    ms_bitmap = CreateCompatibleBitmap(dc, w, h);
    ReleaseDC(NULL, dc);
    depth = wxDisplayDepth();
  }
  if (ms_bitmap)
    ok = TRUE;
  else
    ok = FALSE;
}

wxBitmap::wxBitmap(char *bitmap_file):
  wxbBitmap(bitmap_file)
{
  ms_bitmap = LoadBitmap(wxhInstance, bitmap_file);
  if (ms_bitmap)
  {
    ok = TRUE;
    BITMAP bm;
    GetObject(ms_bitmap, sizeof(BITMAP), (LPSTR) &bm);
    width = bm.bmWidth;
    height = bm.bmHeight;
    depth = bm.bmPlanes;
  }
  else
  {
    ok = FALSE;
    width = 0;
    height = 0;
    depth = 0;
  }

  wxTheBitmapList->Append(this);
}

wxBitmap::~wxBitmap(void)
{
  if (ms_bitmap)
    DeleteObject(ms_bitmap);

  wxTheBitmapList->DeleteObject(this);
}
