/*
 * File:     wx_mf.cc
 * Purpose:  Metafiles and metafile DCs
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "wx.h"
#pragma hdrstop
#include "wx_privt.h"

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/*
#include <windows.h>
#include "common.h"
#include "wx_frame.h"
#include "wx_dc.h"
#include "wx_mf.h"
#include "wx_utils.h"
#include "wx_canvs.h"
#include "wx_dialg.h"
#include "wx_main.h"
*/

/*
 * Metafiles - Windows 3.1 only
 * Currently, the only purpose for making a metafile is to put
 * it on the clipboard.
 */

wxMetaFile::wxMetaFile(void)
{
  metafile = NULL;
}

wxMetaFile::~wxMetaFile(void)
{
  if (metafile)
    { DeleteMetaFile(metafile); metafile = NULL; }
}

Bool wxMetaFile::SetClipboard(int width, int height)
{
  HANDLE data = GlobalAlloc(GHND, sizeof(METAFILEPICT) + 1);
  METAFILEPICT *mf = (METAFILEPICT *)GlobalLock(data);

  mf->mm = MM_TEXT;
  mf->xExt = width;
  mf->yExt = height;
  mf->hMF = metafile;
  GlobalUnlock(data);
  metafile = NULL;

  wxFrame *frame = wxTheApp->wx_frame;
  if (!frame)
    return FALSE;

  wxWnd *wnd = (wxWnd *)frame->handle;
  OpenClipboard(wnd->handle);
  EmptyClipboard();
  HANDLE success = SetClipboardData(CF_METAFILEPICT, data);
  CloseClipboard();
  return (Bool)success;
}

/*
 * Metafile device context
 *
 */

wxMetaFileDC::wxMetaFileDC(char *file)
{
  old_bitmap = NULL;
  old_pen = NULL;
  old_brush = NULL;
  old_font = NULL;

  metafile = NULL;
  wx_interactive = FALSE;
  canvas = NULL;
  font = NULL;

  device = wxDEVICE_WINDOWS;

  cdc = NULL;
  dont_delete = FALSE;
  clipping = FALSE;

  logical_origin_x = 0;
  logical_origin_y = 0;

  device_origin_x = 0;
  device_origin_y = 0;

  logical_scale_x = 1.0;
  logical_scale_y = 1.0;

  user_scale_x = 1.0;
  user_scale_y = 1.0;

  mapping_mode = MM_TEXT;

  min_x = 1000.0;
  min_y = 1000.0;
  max_x = -1000.0;
  max_y = -1000.0;
  title = NULL;

  filename = NULL;

  cdc = CreateMetaFile(file);
  ok = TRUE;
  SetMapMode(MM_TEXT);

  current_logical_function = -1;
  current_pen = wxBLACK_PEN;
  current_brush = wxBLACK_BRUSH;
  current_background_brush = wxWHITE_BRUSH;
  current_text_foreground = *wxBLACK;
  current_text_background = *wxWHITE;
  Colour = wxColourDisplay();
}

wxMetaFileDC::~wxMetaFileDC(void)
{
}

wxMetaFile *wxMetaFileDC::Close(void)
{
  HANDLE mf = CloseMetaFile(cdc);
  if (mf)
  {
    wxMetaFile *wx_mf = new wxMetaFile;
    wx_mf->metafile = mf;
    return wx_mf;
  }
  return NULL;
}

void wxMetaFileDC::SetMapMode(int mode)
{
  mapping_mode = mode;

//  int pixel_width = 0;
//  int pixel_height = 0;
//  int mm_width = 0;
//  int mm_height = 0;

  float mm2pixelsX = 10.0;
  float mm2pixelsY = 10.0;

  switch (mode)
  {
    case MM_TWIPS:
    {
      logical_scale_x = (float)(twips2mm * mm2pixelsX);
      logical_scale_y = (float)(twips2mm * mm2pixelsY);
      break;
    }
    case MM_POINTS:
    {
      logical_scale_x = (float)(pt2mm * mm2pixelsX);
      logical_scale_y = (float)(pt2mm * mm2pixelsY);
      break;
    }
    case MM_METRIC:
    {
      logical_scale_x = mm2pixelsX;
      logical_scale_y = mm2pixelsY;
      break;
    }
    case MM_LOMETRIC:
    {
      logical_scale_x = (float)(mm2pixelsX/10.0);
      logical_scale_y = (float)(mm2pixelsY/10.0);
      break;
    }
    default:
    case MM_TEXT:
    {
      logical_scale_x = 1.0;
      logical_scale_y = 1.0;
      break;
    }
  }
//  SetViewportExtEx(cdc, VIEWPORT_EXTENT, VIEWPORT_EXTENT, NULL);
  window_ext_x = MS_XDEV2LOG(VIEWPORT_EXTENT);
  window_ext_y = MS_YDEV2LOG(VIEWPORT_EXTENT);
//  SetWindowExtEx(cdc, window_ext_x, window_ext_y, NULL);
}
