/*
 * File:     wx_panel.cc
 * Purpose:  wxPanel class implementation (X version)
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <iostream.h>

#include "common.h"
#include "wx_panel.h"
#include "wx_frame.h"
#include "wx_utils.h"
#include "wx_privt.h"

#ifdef wx_motif
#include <Xm/Form.h>
#include <Xm/BulletinB.h>
#include <Xm/PushB.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#endif

#ifdef wx_xview
#include <xview/panel.h>
#include <xview/scrollbar.h>
#include <xview/dragdrop.h>
extern void wxPanelEventProc(Panel panel, Event *event); // See wx_item.cc
#endif

// Constructor
wxPanel::wxPanel(void)
{
  new_line = FALSE;
  label_position = wxHORIZONTAL;

  hSpacing = PANEL_HSPACING;
  vSpacing = PANEL_VSPACING;
  initial_hspacing = hSpacing ;
  initial_vspacing = vSpacing ;
  current_hspacing = hSpacing ;
  current_vspacing = vSpacing ;

  labelFont = NULL ;
  buttonFont = NULL ;
  backColour = NULL ;
  labelColour = NULL ;
  buttonColour = NULL ;

#ifdef wx_motif
  // For absolute-positioning auto layout
  cursor_x = PANEL_LEFT_MARGIN;
  cursor_y = PANEL_TOP_MARGIN;
  max_height = 0;
  max_line_height = 0;
  max_width = 0;
  last_created = 0 ;
  borderWidget = 0;
  new_line = TRUE;
  firstRowWidget = NULL;
  lastWidget = NULL;
  panelWidget = NULL;
  allRelative = TRUE;
  absolutePositioning = FALSE;
  currentRow = 0;
  currentCol = 0;
  handle = NULL;
#endif
#ifdef wx_xview
  x_create = 0 ;
  y_create = 0 ;
#endif
  has_child = FALSE ;
}

wxPanel::wxPanel(wxFrame *frame, int x, int y, int width, int height,
                 int style, char *name):
  wxbPanel(frame, x, y, width, height, style, name)
{
  Create(frame, x, y, width, height, style, name);
}

wxPanel::wxPanel(wxPanel *panel, int x, int y, int width, int height,
                 int style,char *name):
  wxbPanel(panel, x, y, width, height, style, name)
{
#ifdef wx_motif
  windowName = name ;
#endif
  Create(panel, x, y, width, height, style, name);
}

Bool wxPanel::Create(wxFrame *frame, int x, int y, int width, int height,
                     int style, char *name)
{
  new_line = FALSE;
  label_position = wxHORIZONTAL;

  if (!frame)
    return FALSE;

  window_parent = frame;

  has_child = FALSE ;

  hSpacing = PANEL_HSPACING;
  vSpacing = PANEL_VSPACING;
  
  initial_hspacing = hSpacing ;
  initial_vspacing = vSpacing ;

  current_hspacing = hSpacing ;
  current_vspacing = vSpacing ;

  labelFont = NULL ;
  buttonFont = NULL ;
  backColour = NULL ;
  labelColour = NULL ;
  buttonColour = NULL ;

#ifdef wx_motif
  windowName = copystring(name);

  // For absolute-positioning auto layout
  cursor_x = PANEL_LEFT_MARGIN;
  cursor_y = PANEL_TOP_MARGIN;
  max_height = 0;
  max_line_height = 0;
  max_width = 0;

  borderWidget = 0;

  firstRowWidget = NULL;
  lastWidget = NULL;
  allRelative = TRUE;
  absolutePositioning = FALSE;
  currentRow = 0;
  currentCol = 0;

  if (style & wxABSOLUTE_POSITIONING)
    absolutePositioning = TRUE;
  else
    new_line = TRUE;
  last_created = 0 ;

  if (style & wxBORDER)
  {
    borderWidget = XtVaCreateManagedWidget("panelBorder",
                    xmFrameWidgetClass,      frame->clientArea,
                    XmNshadowType,           XmSHADOW_IN,
                    XmNmarginHeight,         0,
                    XmNmarginWidth,          0,
                    NULL);
  }

  panelWidget = XtVaCreateManagedWidget(windowName,
                  absolutePositioning ? xmBulletinBoardWidgetClass: xmFormWidgetClass,
                    borderWidget ? borderWidget : frame->clientArea,
/*
                  XmNresizePolicy, XmRESIZE_NONE,
                  XmNmarginHeight, 0,
                  XmNmarginWidth, 0,
*/
                  NULL);

  if (absolutePositioning)
    XtVaSetValues(panelWidget,
                  XmNresizePolicy, XmRESIZE_NONE,
                  NULL) ;

  if (style & wxFLAT)
    XtVaSetValues(panelWidget,
                  XmNborderWidth,1,
                  NULL) ;

  handle = (char *)panelWidget;

  wxWidgetHashTable->Put((long)panelWidget, this);

  XtTranslations ptr;
  XtOverrideTranslations(panelWidget,
              ptr = XtParseTranslationTable("<Configure>: resize()"));

  XtFree((char *)ptr);

  SetSize(x, y, width, height);
#endif
#ifdef wx_xview
  int real_y = frame->y_offset;
  if (y > -1)
    real_y = y + frame->y_offset;  // Allow for possible menu bar

  Bool scrollable = style & wxVSCROLL;

  Frame x_frame = (Frame)(frame->GetHandle());
  Panel x_panel = (Panel)xv_create(x_frame, scrollable ? SCROLLABLE_PANEL : PANEL,
                                   PANEL_BACKGROUND_PROC, wxPanelEventProc,
                                   PANEL_ACCEPT_KEYSTROKE, TRUE,
                                   WIN_CLIENT_DATA, (char *)this,
                                   PANEL_LAYOUT, PANEL_HORIZONTAL, XV_SHOW, FALSE,
                                   NULL);
  Scrollbar sb; 
  if (scrollable)
    sb = xv_create(x_panel, SCROLLBAR,
                   NULL);

  if (x > -1)
  {
    xv_set(x_panel, XV_X, x, NULL);
    x_create = x;
  }
  else
    x_create = 0;

  if (y > -1)
    xv_set(x_panel, XV_Y, real_y, NULL);

  y_create = real_y;

  // If width = -1, will extend to edge of frame
  xv_set(x_panel, XV_WIDTH, width, NULL);

  if (height > -1)
    xv_set(x_panel, XV_HEIGHT, height, NULL);

  xv_set(x_panel, XV_SHOW, TRUE, NULL);

  handle = (char *)x_panel;
#endif
  if (frame) frame->AddChild(this);
  return TRUE;
}

Bool wxPanel::Create(wxPanel *frame, int x, int y, int width, int height, int style, char *name)
{
  if (frame)
  {
#ifdef wx_motif
    // mixing panels with differents modes doesn't seem to work, so
    // set style to parent's style.
    if (frame->absolutePositioning)
      style |= wxABSOLUTE_POSITIONING ;
    else
      style &= ~wxABSOLUTE_POSITIONING ;
#endif
    frame->AddChild(this);
  }

  new_line = FALSE;
  label_position = wxHORIZONTAL;

  window_parent = frame;

  has_child = FALSE ;

  hSpacing = PANEL_HSPACING;
  vSpacing = PANEL_VSPACING;

  initial_hspacing = hSpacing ;
  initial_vspacing = vSpacing ;

  current_hspacing = hSpacing ;
  current_vspacing = vSpacing ;

  labelFont = frame->labelFont ;
  buttonFont = frame->buttonFont ;
  backColour = frame->backColour ;
  labelColour = frame->labelColour ;
  buttonColour = frame->buttonColour ;

#ifdef wx_motif
  borderWidget = 0;
  firstRowWidget = NULL;
  lastWidget = NULL;
  allRelative = TRUE;
  absolutePositioning = FALSE;
  currentRow = 0;
  currentCol = 0;

  if (style & wxABSOLUTE_POSITIONING)
    absolutePositioning = TRUE;
  else
    new_line = TRUE;
  last_created = 0 ;

  if (style & wxBORDER)
  {
    borderWidget = XtVaCreateManagedWidget("main_window",
                    xmFrameWidgetClass,      frame->panelWidget,
                    XmNshadowType,           XmSHADOW_IN,
                    XmNmarginHeight,        0,
                    XmNmarginWidth,         0,
                    NULL);
  }
  else if (style & wxFLAT)
  {
    borderWidget = XtVaCreateManagedWidget("main_window",
                    xmRowColumnWidgetClass, frame->panelWidget,
                    XmNpacking,             XmPACK_NONE,
                    XmNmarginHeight,        0,
                    XmNmarginWidth,         0,
                    XmNborderWidth,         1,
                    NULL);
  }

  panelWidget = XtVaCreateManagedWidget(windowName,
                  absolutePositioning ? xmBulletinBoardWidgetClass: xmFormWidgetClass,
                    borderWidget ? borderWidget : frame->panelWidget,
                  NULL);

  if (absolutePositioning)
    XtVaSetValues(panelWidget,
                  XmNresizePolicy, XmRESIZE_NONE,
                  NULL) ;

  handle = (char *)panelWidget;


  if (frame->absolutePositioning)
    frame->AdvanceCursor(this) ;
  AttachWidget(frame,borderWidget?borderWidget:panelWidget,x,y,width,height) ;
  //SetSize(x, y, width, height);
#endif
#ifdef wx_xview

  // Compute my creation pos.
  wxPanel *father = (wxPanel*)GetParent();
  father->GetCursor(&x_create,&y_create);
  if (y > -1)
    y_create += y ;
  if (x > -1)
    x_create += x ;

  int x_orig,y_orig;
  x_orig = x_create ;
  y_orig = y_create ;
  while(!wxSubType(father->__type,wxTYPE_FRAME))
  {
    x_orig += father->x_create ;
    y_orig += father->y_create ;
    father = (wxPanel*)(father->GetParent()) ;
  }

  Bool scrollable = style & wxVSCROLL;

  Frame x_frame = (Frame)(father->GetHandle());
  Panel x_panel = (Panel)xv_create(x_frame, scrollable ? SCROLLABLE_PANEL : PANEL,
                                   PANEL_BACKGROUND_PROC, wxPanelEventProc,
                                   PANEL_ACCEPT_KEYSTROKE, TRUE,
                                   WIN_CLIENT_DATA, (char *)this,
                                   PANEL_LAYOUT, PANEL_HORIZONTAL, XV_SHOW, FALSE,
                                   NULL);
  Scrollbar sb; 
  if (scrollable)
    sb = xv_create(x_panel, SCROLLBAR,
                   NULL);

  xv_set(x_panel, XV_X, x_orig, NULL);

  xv_set(x_panel, XV_Y, y_orig, NULL);

  // If width = -1, will extend to edge of frame
  xv_set(x_panel, XV_WIDTH, width, NULL);

  if (height > -1)
    xv_set(x_panel, XV_HEIGHT, height, NULL);

  xv_set(x_panel, XV_SHOW, TRUE, NULL);

  handle = (char *)x_panel;
#endif
  return TRUE;

}


wxPanel::~wxPanel(void)
{
#ifdef wx_motif
  if (borderWidget)
    XtDestroyWidget(borderWidget);
#endif
}

void wxPanel::SetSize(int x, int y, int w, int h)
{
#ifdef wx_motif
  Widget ancestor = borderWidget? borderWidget:panelWidget ;
//  XtUnmanageChild(borderWidget ? borderWidget : panelWidget);

  if (x > -1)
  {
    if (XtIsSubclass(ancestor,xmFormWidgetClass))
      XtVaSetValues(ancestor,
                    XmNx, x, 
                    XmNleftAttachment, XmATTACH_SELF,
                    NULL);
    else
      XtVaSetValues(ancestor,
                    XmNx, x, 
                    NULL);

/*
    XtVaSetValues(borderWidget ? borderWidget : panelWidget,
                  XmNx, x, 
                  NULL);
*/
  }
  if (y > -1)
  {
    if (XtIsSubclass(ancestor,xmFormWidgetClass))
      XtVaSetValues(ancestor,
                    XmNy, y,
                    XmNtopAttachment, XmATTACH_SELF,
                    NULL);
    else
      XtVaSetValues(ancestor,
                    XmNy, y,
                    NULL);

/*
    XtVaSetValues(borderWidget ? borderWidget : panelWidget,
                  XmNy, y,
                  NULL);
*/
  }

  if (w > -1)
  {
    // I can't remember why, but order of setting is important!
    if (borderWidget)
      XtVaSetValues(borderWidget, XmNwidth, w, NULL);
    XtVaSetValues(panelWidget, XmNwidth, w, NULL);
  }
//  Dimension yy1,yy2;
  if (h > -1)
  {
    // I can't remember why, but order of setting is important!
    if (borderWidget)
      XtVaSetValues(borderWidget, XmNheight, h, NULL);

    // Test
//    XtVaGetValues(panelWidget, XmNheight, &yy2, NULL);

    XtVaSetValues(panelWidget, XmNheight, h, NULL);

    // Test
//    XtVaGetValues(panelWidget, XmNheight, &yy1, NULL);
  }
//  XtManageChild(borderWidget ? borderWidget : panelWidget);
  // Again, can't remember why, but this call must be done.
  wxWindow::SetClientSize(w,h) ;
//  XtVaGetValues(panelWidget, XmNheight, &yy1, NULL);
#endif
#ifdef wx_xview
  if (x == -1 || y == -1)
  {
    int xx,yy ;
    GetPosition(&xx, &yy);
    if (x == -1) x = xx ;
    if (y == -1) y = yy ;
  }

  if (w == -1 || h== -1)
  {
    int ww,hh ;
    GetSize(&ww, &hh);
    if (w == -1) w = ww ;
    if (h == -1) h = hh ;
  }

  int real_y = y;

  if (window_parent && wxSubType(window_parent->__type, wxTYPE_FRAME))
    real_y += ((wxFrame *)window_parent)->y_offset;  // Allow for possible menu bar

  Xv_opaque object = (Xv_opaque)handle;

  (void)xv_set(object, XV_X, x, XV_Y, real_y, NULL);

  (void)xv_set(object, XV_WIDTH, w, XV_HEIGHT, h, NULL);
#endif
  OnSize(w, h);
}

void wxPanel::SetClientSize(int w, int h)
{
#ifdef wx_motif
  SetSize(-1, -1, w, h);
#endif
#ifdef wx_xview
  wxWindow::SetClientSize(w, h);
#endif
}

void wxPanel::GetPosition(int *x, int *y)
{
#ifdef wx_motif
  Dimension xx, yy;
  XtVaGetValues(borderWidget ? borderWidget : (Widget)handle,
                XmNx, &xx, XmNy, &yy, NULL);
  *x = xx;
  *y = yy;
#else
  wxWindow::GetPosition(x, y);

  if (window_parent && wxSubType(window_parent->__type, wxTYPE_FRAME))
    *y = *y - ((wxFrame *)window_parent)->y_offset;  // Allow for menu bar
#endif
}

/*****************************************************************
 * ITEM PLACEMENT FUNCTIONS
 *****************************************************************/


// Start a new line
#ifdef wx_motif
void wxPanel::RealNewLine(void)
{
  if (!absolutePositioning)
    return ;
  
  //cursor_x = PANEL_LEFT_MARGIN;
  cursor_x = initial_hspacing;
  if (max_line_height == 0)
  {
    cursor_y += current_vspacing;
  }
  else
    cursor_y += current_vspacing + max_line_height;
  max_line_height = 0;
  new_line = FALSE ;
}
#endif

void wxPanel::NewLine(void)
{
#ifdef wx_motif
  if (new_line)
    current_vspacing += vSpacing ;
  else
    current_vspacing = vSpacing ;
  new_line = TRUE ;
#endif
#ifdef wx_xview
  Panel panel = (Panel)handle;
  if (new_line)
  {
      (void) xv_create(panel, PANEL_MESSAGE,
                             PANEL_LAYOUT, PANEL_HORIZONTAL,
                             PANEL_NEXT_ROW, vSpacing, NULL);
  }
  new_line = TRUE;
#endif
}

void wxPanel::NewLine(int pixels)
{
#ifdef wx_motif
  if (new_line)
    current_vspacing += pixels ;
  else
    current_vspacing = pixels ;
  new_line = TRUE ;
#endif
#ifdef wx_xview
  Panel panel = (Panel)handle;
  if (new_line)
  {
      (void) xv_create(panel, PANEL_MESSAGE,
                             PANEL_LAYOUT, PANEL_HORIZONTAL,
                             PANEL_NEXT_ROW, pixels, NULL);
  }
  new_line = TRUE;
#endif
}

void wxPanel::Tab(void)
{
#ifdef wx_motif
  if (!absolutePositioning)
    current_hspacing += hSpacing ;
  else
  {
    // If we're doing a newline on a bulletin board, we work this
    // out ourselves.
    cursor_x += hSpacing;
    if (cursor_x > max_width)
      max_width = cursor_x;
  }
#endif
#ifdef wx_xview
  current_hspacing += hSpacing ;
  Panel panel = (Panel)handle;
  (void)xv_set(panel, PANEL_ITEM_X_GAP, current_hspacing, NULL);
#endif
}

void wxPanel::Tab(int pixels)
{
#ifdef wx_motif
  if (!absolutePositioning)
    current_hspacing += pixels ;
  else
  {
    // If we're doing a newline on a bulletin board, we work this
    // out ourselves.
    cursor_x += pixels;
    if (cursor_x > max_width)
      max_width = cursor_x;
  }
#endif
#ifdef wx_xview
  current_hspacing += pixels ;
  Panel panel = (Panel)handle;

/*
    (void) xv_create(panel, PANEL_MESSAGE,
                           PANEL_LAYOUT, PANEL_HORIZONTAL,
                           XV_WIDTH, pixels,
                           NULL);
*/
  (void)xv_set(panel, PANEL_ITEM_X_GAP, current_hspacing, NULL);
#endif
}

void wxPanel::GetCursor(int *x, int *y)
{
#ifdef wx_motif
  *x = cursor_x;
  *y = cursor_y;
#endif
#ifdef wx_xview
  Xv_opaque object = (Xv_opaque)handle;
  *x = (int)xv_get(object, PANEL_ITEM_X);
  *y = (int)xv_get(object, PANEL_ITEM_Y);
#endif
}

// Set/get horizontal spacing
void wxPanel::SetHorizontalSpacing(int sp)
{
#ifdef wx_motif
  hSpacing = sp;
  current_hspacing = sp;
/*
  XtVaSetValues((Widget)handle,
                  XmNmarginWidth, sp,
                  NULL);
*/
#endif
#ifdef wx_xview
  Xv_opaque object = (Xv_opaque)handle;
  (void)xv_set(object, PANEL_ITEM_X_GAP, sp, NULL);
#endif
}

int wxPanel::GetHorizontalSpacing(void)
{
#ifdef wx_motif
  return hSpacing;
#endif
#ifdef wx_xview
  Xv_opaque object = (Xv_opaque)handle;
  return (int)xv_get(object, PANEL_ITEM_X_GAP);
#endif
}

// Set/get vertical spacing
void wxPanel::SetVerticalSpacing(int sp)
{
#ifdef wx_motif
  vSpacing = sp;
  current_vspacing = sp;
/*
  XtVaSetValues((Widget)handle,
                  XmNmarginHeight, sp,
                  NULL);
*/
#endif
#ifdef wx_xview
  Xv_opaque object = (Xv_opaque)handle;
  (void)xv_set(object, PANEL_ITEM_Y_GAP, sp, NULL);
#endif
}

int wxPanel::GetVerticalSpacing(void)
{
#ifdef wx_motif
  return vSpacing;
#endif
#ifdef wx_xview
  Xv_opaque object = (Xv_opaque)handle;
  return (int)xv_get(object, PANEL_ITEM_Y_GAP);
#endif
}

// Fits the panel around the items
void wxPanel::Fit(void)
{
#ifdef wx_motif
  if (absolutePositioning)
  {
    RealAdvanceCursor() ;
    SetClientSize(max_width+initial_hspacing,max_height+initial_vspacing);
  }
  else
  {
    int maxX = 0;
    int maxY = 0;
    wxNode *node = GetChildren()->First();
    while (node)
    {
      wxWindow *win = (wxWindow *)node->Data();
      int x, y, w, h;
      win->GetPosition(&x, &y);
      win->GetSize(&w, &h);
      if ((x + w) > maxX)
        maxX = x + w;
      if ((y + h) > maxY)
        maxY = y + h;
      node = node->Next();
    }
    SetClientSize(maxX + initial_hspacing, maxY + initial_vspacing);
    OptimizeLayout();
  }
#endif
#ifdef wx_xview
//  Xv_opaque object = (Xv_opaque)handle;
  // Pre-fit
//  window_fit(object); // Doesn't seem to work always
  int maxX = 0;
  int maxY = 0;
  wxNode *node = GetChildren()->First();
  while (node)
  {
    wxWindow *win = (wxWindow *)node->Data();
    if ((wxSubType(win->__type, wxTYPE_PANEL) &&
         !wxSubType(win->__type, wxTYPE_DIALOG_BOX)) ||
        wxSubType(win->__type, wxTYPE_TEXT_WINDOW) ||
        wxSubType(win->__type, wxTYPE_CANVAS))
    {
      int x, y, w, h;
      win->GetPosition(&x, &y);
      win->GetSize(&w, &h);
      if ((x + w) > maxX)
        maxX = x + w;
      if ((y + h) > maxY)
        maxY = y + h;
    }
    node = node->Next();
  }

  // Add subpanels
  wxNode *child = children->First() ;
//  int ww,hh ;
//  GetSize(&ww,&hh) ;
  while(child)
  {
    wxPanel *panel = (wxPanel*)child->Data() ;
    child = child->Next() ;
    if (!wxSubType(panel->__type, wxTYPE_PANEL))
      continue ;
    panel->Fit() ;
    int cw,ch;
    GetSize(&cw,&ch) ;
    cw += panel->x_create ;
    ch += panel->y_create ;
    if (cw>maxX)
      maxX = cw ;
    if (ch>maxY)
      maxY = ch ;
  }
  SetClientSize(maxX, maxY);
#endif
}

#ifdef wx_motif
// Update next cursor position
void wxPanel::RealAdvanceCursor(void)
{
  wxWindow *item = last_created ;
  if (item)
  {
    int width, height;
    int x, y;
    item->GetSize(&width, &height);
    item->GetPosition(&x, &y);

    if ((x + width) > max_width)
      max_width = x + width;
    if ((y + height) > max_height)
      max_height = y + height;
    if (height > max_line_height)
      max_line_height = height;
/*
fprintf(stderr,"maxwidth %d maxheight %d maklineheight %d\n",
        max_width,max_height,max_line_height);
fprintf(stderr,"Newline %d\n",new_line) ;
*/
    cursor_x = x + width + current_hspacing;
    cursor_y = y;
    last_created = NULL ;
  }
  if (new_line)
    RealNewLine() ;
}
#endif

// Update next cursor position
void wxPanel::AdvanceCursor(wxWindow *item)
{
#ifdef wx_motif
  // For calculating positions on a bulletin board.
/*
  int width, height;
  int x, y;
  item->GetSize(&width, &height);
  item->GetPosition(&x, &y);

  if ((x + width) > max_width)
    max_width = x + width;
  if ((y + height) > max_height)
    max_height = y + height;
  if (height > max_line_height)
    max_line_height = height;

  cursor_x = x + width + hSpacing;
  cursor_y = y;
*/
  last_created = item ;
#endif
#ifdef wx_xview
#endif
}

// If x or y are not specified (i.e. < 0), supply
// values based on left to right, top to bottom layout.
// Internal use only.
void wxPanel::GetValidPosition(int *x, int *y)
{
#ifdef wx_motif
  // For calculating positions on a bulletin board.
  if (*x < 0)
    *x = cursor_x;

  if (*y < 0)
    *y = cursor_y;
#endif
#ifdef wx_xview
#endif
}

#ifdef wx_xview
void wxPanel::DragAcceptFiles(Bool accept)
{

/*
  if (accept)
  {
    if (dropSite)
      xv_destroy(dropSite);

    Xv_Window pw = xv_get((Panel)handle, PANEL_FIRST_PAINT_WINDOW, NULL);
    dropSite = xv_create(pw, DROP_SITE_ITEM,
                           DROP_SITE_ID, NewId(),
                           DROP_SITE_REGION, xv_get(pw, WIN_RECT),
                           DROP_SITE_EVENT_MASK, DND_ENTERLEAVE,
                           NULL);
  }
  else if (dropSite) xv_destroy(dropSite);
*/
}
#endif


#ifdef wx_motif
void wxPanel::AttachWidget(wxPanel *panel, Widget formWidget,
                          int x, int y, int width, int height)
{

//fprintf(stderr,"Subpanel %x Panel %x\n",this,panel) ;
//fprintf(stderr,"Panel: ih %d iv %d ch %d cv %d hs %d vs %d\n",
//	panel->initial_hspacing,panel->initial_vspacing,
//	panel->current_hspacing,panel->current_vspacing,
//	panel->hSpacing,panel->vSpacing) ;

  if ((x > -1) || (y > -1))
    panel->allRelative = FALSE;

  //itemOrientation = panel->label_position;

  wxWidgetHashTable->Put((long)formWidget, this);
  XtTranslations ptr;
  XtOverrideTranslations(formWidget,
              ptr = XtParseTranslationTable("<Configure>: resize()"));
  XtFree((char *)ptr);

  if (!panel->absolutePositioning)
  {
    XtVaSetValues(formWidget,
      XmNleftOffset, panel->current_hspacing,
      NULL);

    if (panel->new_line)
    {
      panel->currentRow ++;
      panel->currentCol = 0;

      if (x == -1)
        XtVaSetValues(formWidget,
                      XmNleftAttachment, XmATTACH_FORM,
                      XmNleftOffset,     panel->initial_hspacing,
                      NULL);
      if (y == -1)
        XtVaSetValues(formWidget,
                      XmNtopAttachment,  panel->firstRowWidget ? XmATTACH_WIDGET: XmATTACH_FORM,
                      XmNtopWidget,      panel->firstRowWidget,
                      XmNtopOffset,      panel->current_vspacing,
                      NULL);

      panel->lastWidget = formWidget;
      panel->firstRowWidget = formWidget;
      panel->new_line = FALSE;
    }
    else
    {
      panel->currentCol ++;
      if (x == -1)
        XtVaSetValues(formWidget,
                      XmNleftAttachment, panel->lastWidget ? XmATTACH_WIDGET: XmATTACH_FORM,
                      XmNleftWidget,     panel->lastWidget,
                      XmNleftOffset,     panel->lastWidget? panel->current_hspacing:panel->initial_hspacing,
                      NULL);
      if (y == -1)
        XtVaSetValues(formWidget,
                      XmNtopAttachment,  panel->lastWidget ? XmATTACH_OPPOSITE_WIDGET: XmATTACH_FORM,
                      XmNtopWidget,      panel->lastWidget,
                      XmNtopOffset,      panel->lastWidget? 0:panel->initial_vspacing,
                      NULL);
      // The following constraint is meant to help buttons align neatly in case one
      // is the default and is bigger than the rest. Unfortunately it seems to have the effect
      // of shrinking the current/previous widget and maybe some other nasty things.
  /*
      if (height == -1)
        XtVaSetValues(formWidget,
                      XmNbottomAttachment,  panel->lastWidget ? XmATTACH_OPPOSITE_WIDGET: XmATTACH_FORM,
                      XmNbottomWidget,      panel->lastWidget,
  		    NULL);
  */

      panel->lastWidget = formWidget;
    }

    //rowNumber = panel->currentRow;
    //colNumber = panel->currentCol;
  }


  SetSize(x, y, width, height);

  if (panel->absolutePositioning)
    return;

  if (!panel->firstRowWidget)
    panel->firstRowWidget = formWidget;
  else
  {
    Dimension hh1, hh2;
    XtVaGetValues(formWidget, XmNheight, &hh1, NULL);
    XtVaGetValues(panel->firstRowWidget, XmNheight, &hh2, NULL);

    if (hh1 > hh2)
      panel->firstRowWidget = formWidget;
  }

  current_hspacing = hSpacing ;
  current_vspacing = vSpacing ;

}
#endif

void wxPanel::Centre(int direction)
{
  int x, y, width, height, panel_width, panel_height, new_x, new_y;

  wxPanel *father = (wxPanel *)GetParent();
  if (!father)
    return;

  father->GetClientSize(&panel_width, &panel_height);
  GetSize(&width, &height);
  GetPosition(&x, &y);

  new_x = -1;
  new_y = -1;

  if (direction == wxHORIZONTAL || direction == wxBOTH)
    new_x = (int)((panel_width - width)/2);

  if (direction == wxVERTICAL || direction == wxBOTH)
    new_y = (int)((panel_height - height)/2);

  SetSize(new_x, new_y, -1, -1);

}

void wxPanel::Enable(Bool Flag)
{
#ifdef wx_motif
  if (panelWidget)
    XtSetSensitive(panelWidget,Flag) ;
#endif
}

void wxPanel::AddChild(wxObject *child)
{

  if (!has_child)
  {
    initial_hspacing = hSpacing ;
    initial_vspacing = vSpacing ;
#ifdef wx_motif
/*
    if (borderWidget)
    {
      XtVaSetValues(borderWidget,
                XmNmarginHeight,initial_vspacing,
                XmNmarginWidth,initial_hspacing,
                    NULL) ;
      initial_hspacing = 0 ;
      initial_vspacing = 0 ;
    }
*/
#endif
  }
  has_child = TRUE ;

#ifdef wx_motif
  if (absolutePositioning)
  {
    cursor_x = hSpacing ;
    cursor_y = vSpacing ;
    RealAdvanceCursor() ;
  }
#endif
  current_hspacing = hSpacing ;
  current_vspacing = vSpacing ;

  children->Append(child) ;

}
