/*
 * File:     wx_timer.cc
 * Purpose:  wxTimer implementation (X version)
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <time.h>
#include <sys/types.h>
#ifdef __sgi
#include <sys/time.h>
#else
#include <sys/timeb.h>
#endif
#include "wx_timer.h"
#include "wx_list.h"
#include "wx_main.h"
#include "wx_frame.h"
#include "wx_utils.h"

#ifdef wx_motif
#include <Xm/Xm.h>
#endif

// Can't find an include file with this in!!
extern "C" int ftime(struct timeb *);

#ifdef wx_motif
void wxTimerCallback(wxTimer *timer)
{
  if (!timer->oneShot)
    timer->timerId = XtAppAddTimeOut(wxTheApp->appContext, timer->milli, 
                                     (XtTimerCallbackProc) wxTimerCallback, (XtPointer) timer);
  else
    timer->timerId = 0;
  timer->Notify();
}
#endif

#ifdef wx_xview
Notify_value wxTimerFunc(Notify_client client, int which);
#endif

wxTimer::wxTimer(void)
{
#ifdef wx_motif
  timerId = 0;
  milli = 0;
#endif
#ifdef wx_xview
  timerval.it_value.tv_usec = 0;
  timerval.it_interval.tv_usec = 0;
#endif
}

wxTimer::~wxTimer(void)
{
  Stop();
}

Bool wxTimer::Start(int milliseconds,Bool mode)
{
  oneShot = mode ;
#ifdef wx_motif
  milli = milliseconds;
  timerId = XtAppAddTimeOut(wxTheApp->appContext, milliseconds,
                               (XtTimerCallbackProc) wxTimerCallback, (XtPointer) this);
  return TRUE;
#endif
#ifdef wx_xview
  if (milliseconds > -1)
  {
    long secs = (long)(milliseconds/1000);
    long usecs = (long)(milliseconds - 1000*secs)*1000;

    timerval.it_value.tv_sec = secs;
    timerval.it_interval.tv_sec = secs;

    timerval.it_value.tv_usec = usecs;
    timerval.it_interval.tv_usec = usecs;
  }

  notify_set_itimer_func((Notify_client)this, (Notify_func)wxTimerFunc, ITIMER_REAL,
                         &timerval, NULL);
  return TRUE;
#endif
}

void wxTimer::Stop(void)
{
#ifdef wx_motif
  if (timerId > 0)
  {
    XtRemoveTimeOut(timerId);
    timerId = 0;
  }
#endif
#ifdef wx_xview
  notify_set_itimer_func((Notify_client)this, NOTIFY_FUNC_NULL, ITIMER_REAL,
                         NULL, NULL);
#endif
}

#ifdef wx_xview
Notify_value wxTimerFunc(Notify_client client, int which)
{
  wxTimer *timer = (wxTimer *)client;
  timer->Notify();
  if (timer->oneShot)
    timer->Stop() ;
  return NOTIFY_DONE;
}
#endif

