/*
 * File:     test.cc
 * Purpose:  PROLOGIO Test
 *
 *                       wxWindows 1.50
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                        Date: 7-9-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "wx.h"
#include "read.h"
#include "iostream.h"

Bool errorHandler(int errorType, char *msg);

PrologDatabase database(errorHandler);

// Define a new application type
class MyApp: public wxApp
{ public:
    wxFrame *OnInit(void);
};

// Define a new frame type
class MyFrame: public wxFrame
{ public:
    MyFrame(wxFrame *frame, char *title, int x, int y, int w, int h);
    void OnMenuCommand(int id);
};

// IDs
#define TEST_QUIT 1
#define TEST_LOAD 2
#define TEST_SAVE 3
#define TEST_CREATE 4

// This statement initializes the whole application and calls OnInit
MyApp myApp;

// `Main program' equivalent, creating windows and returning main app frame
wxFrame *MyApp::OnInit(void)
{
  // Create the main frame window
  MyFrame *frame = new MyFrame(NULL, "PROLOGIO Test", 50, 50, 400, 300);

  // Give it an icon
  frame->SetIcon(new wxIcon("aiai_icn"));

  // Make a menubar
  wxMenu *file_menu = new wxMenu;
  file_menu->Append(TEST_LOAD, "Load");
  file_menu->Append(TEST_SAVE, "Save");
  file_menu->Append(TEST_CREATE, "Create");
  file_menu->Append(TEST_QUIT, "Quit");
  wxMenuBar *menu_bar = new wxMenuBar;
  menu_bar->Append(file_menu, "File");
  frame->SetMenuBar(menu_bar);

  // Make a panel with a message
  (void) new wxPanel(frame, 0, 0, 400, 300);

  // Show the frame
  frame->Show(TRUE);

  // Return the main frame window
  return frame;
}

// My frame constructor
MyFrame::MyFrame(wxFrame *frame, char *title, int x, int y, int w, int h):
  wxFrame(frame, title, x, y, w, h)
{}

// Intercept menu commands
void MyFrame::OnMenuCommand(int id)
{
  switch (id) {
    case TEST_LOAD:
    {
      char *s = wxFileSelector("Load a Prolog file");
      if (s)
      {
        database.Clear();
        database.ReadProlog(s);
      }
      break;
    }
    case TEST_SAVE:
    {
      char *s = wxFileSelector("Save a Prolog file");
      if (s)
      {
        ofstream str(s);
        database.WriteProlog(str);
      }
      break;
    }
    case TEST_CREATE:
    {
      database.Clear();

      PrologExpr *clause1 = new PrologExpr("clause1");
      clause1->AddAttributeValue("id", (long)12345);
      clause1->AddAttributeValueString("name", "fred");
      database.Append(clause1);

      PrologExpr *clause2 = new PrologExpr("clause2");
      clause2->AddAttributeValue("id", (long)654321);
      clause2->AddAttributeValueString("name", "jim");
      database.Append(clause2);

      break;
    }
    case TEST_QUIT:
      delete this;
    break;
  }
}

Bool errorHandler(int errorType, char *msg)
{
  char buf[500];
  sprintf(buf, "Error during loading: %s", msg);
  wxMessageBox(buf, "Error");
  return FALSE;
}
