/*
 * wxclips.cc
 *
 * wxWindows front end to CLIPS
 *
 */

#include <wx.h>

#include <iostream.h>
#include <ctype.h>

#include <wx_help.h>
#include "clipscpp.h"
#include "wx_cmds.h"
#include "wxclips.h"

#include "clips.h"


#ifdef wx_x
#include "aiai.xbm"
#endif

#ifdef wx_msw
char *wxClipsIcon = "door_icn";
#endif

typedef int (*ClipsFunction) ();

#ifdef wx_x
#define COMMAND_PANEL_HEIGHT 130
#define COMMAND_TEXT_WIDTH   350
#endif
#ifdef wx_msw
#define COMMAND_PANEL_HEIGHT 95
#define COMMAND_TEXT_WIDTH   400
#endif

int all_watch = 0;
int compilations_watch = 1;
int facts_watch = 0;
int rules_watch = 0;
int activations_watch = 0;
int statistics_watch = 0;
int globals_watch = 0;
int deffunctions_watch = 0;
int limit = -1;

float wxClipsVersion = 1.10;

wxHelpInstance *HelpInstance = NULL;

wxClipsEnvFrame *wxTheClipsFrame = NULL;

// This statement initialises the whole application
MyApp     myApp;

wxFrame *wxStartApp(void)
{
  wxFrame *main_frame = NULL;
  clipsReturnType = clipsUNKNOWN;
  RouteCommand("(return-result (app-on-init))");

  if (clipsReturnType == clipsLONG)
  {
    main_frame = (wxFrame *)wxGetTypedObject(clipsReturnLong, wxTYPE_FRAME);
  }

  if (!main_frame)
  {
    ClipsError << "Error in wxStartApp: app-on-init should return a main frame ID!\n";
#ifdef wx_x
    if (!wxTheClipsFrame)
      cerr << "Error in wxStartApp: app-on-init should return a main frame ID!\n";
#endif
    return NULL;
  }
  else return main_frame;
}

void wxInitializeClips(void)
{
  // Initialize CLIPS
  wxInitClips();

  AddRunFunction("wxyield", (ClipsFunction)ClipsYield, 100);

  AddRouter("wxrouter", 10,
            (ClipsFunction)FindOutput,
            (ClipsFunction)PrintOutput,
            (ClipsFunction)NULL,
            (ClipsFunction)NULL,
            (ClipsFunction)ExitOutput);

  AddWatchItem("compilations", &compilations_watch, 100);
  AddWatchItem("facts", &facts_watch, 100);
  AddWatchItem("rules", &rules_watch, 100);
  AddWatchItem("activations", &activations_watch, 100);
  AddWatchItem("statistics", &statistics_watch, 100);
  AddWatchItem("globals", &globals_watch, 100);
  AddWatchItem("all", &all_watch, 100);

  SetWatchItem("compilations", compilations_watch);
}

void Usage(void)
{
#ifdef wx_x
  cout << "WXCLIPS command line options:\n";
  cout << "wxclips [-clips    <filename>]  ; clips batch file\n";
  cout << "        [-start]                ; Call (app-on-init), no debugging window\n";
  cout << "        [-h]                    ; this help screen\n";
  cout << "        [-help]                 ; list of generic X window arguments\n";
  exit(0);
#endif
}

// The `main program' equivalent, creating the windows and returning the
// main frame
wxFrame *MyApp::OnInit(void)
{
  wxInitializeClips();

  HelpInstance = new wxHelpInstance(TRUE);
  HelpInstance->Initialize("clipswin");

  char *batchFile = NULL;
  Bool start = FALSE;

  int i = 1;
  while (i < argc)
  {
    if (strcmp(argv[i], "-clips") == 0)
    {
      i ++;
      if (i < argc)
        batchFile = copystring(argv[i]);
      i ++;
    }
    else if (strcmp(argv[i], "-start") == 0)
    {
      i ++;
      start = TRUE;
    }
#ifdef wx_x
    else if (strcmp(argv[i], "-h") == 0)
    {
      i ++;
      Usage();
    }
#endif
    else
    {
#ifdef wx_x
      cerr << "Invalid command-line option.\n";
#endif
      i ++;
      Usage();
    }
  }

  wxFrame *clipsMainFrame = NULL;
  if (batchFile)
  {
    wxExecuteClipsFile(batchFile);
    if (start)
    {
      clipsMainFrame = wxStartApp();
    }
  }

  if (start && clipsMainFrame)
  {
    return clipsMainFrame;
  }
  else
  {
    wxShowClipsWindow();
    return wxTheClipsFrame;
  }
}

void wxShowClipsWindow(void)
{
  if (wxTheClipsFrame)
    wxTheClipsFrame->Iconize(FALSE);
  else
    wxTheClipsFrame = new wxClipsEnvFrame(NULL, "wxCLIPS", 10, 10, 480, 450);
}

void wxDeleteClipsWindow(void)
{
  if (wxTheClipsFrame && wxTheClipsFrame->OnClose())
  {
    delete wxTheClipsFrame;
    wxTheClipsFrame = NULL;
  }
}

// Send a command to CLIPS
void EnterCommand(wxButton& but, wxCommandEvent& event)
{
  char *text = wxTheClipsFrame->command_line->GetValue();
  if (text)
  {
    char *comm = copystring(text);
    wxTheClipsFrame->text_window->WriteText("CLIPS> ");
    wxTheClipsFrame->text_window->WriteText(comm);
    wxTheClipsFrame->text_window->WriteText("\n");
    RouteCommand(comm);
    delete comm;
    PeriodicCleanup(CLIPS_TRUE,CLIPS_FALSE);
    if (wxTheClipsFrame)
    {
      wxTheClipsFrame->command_line->SetValue("");
      wxTheClipsFrame->command_line->SetFocus();
    }
  }
}

// Define my frame constructor
wxClipsEnvFrame::wxClipsEnvFrame(wxFrame *frame, char *title, int x, int y, int w, int h):
  wxFrame(frame, title, x, y, w, h)
{
  panel = NULL;
  text_window = NULL;

  wxTheClipsFrame = this;
  // Give it a status line
//  CreateStatusLine();

  // Give it an icon
#ifdef wx_msw
  wxIcon *icon = new wxIcon(wxClipsIcon);
#endif
#ifdef wx_x
  wxIcon *icon = new wxIcon(aiai_bits, 64, 64);
#endif

  SetIcon(icon);

  // Make a menubar

  wxMenu *file_menu = new wxMenu;

  file_menu->Append(WXCLIPS_BATCH, "&Batch");
  file_menu->Append(WXCLIPS_LOAD, "&Load definitions");
  file_menu->Append(WXCLIPS_SAVE, "&Save definitions");
  file_menu->AppendSeparator();
  file_menu->Append(WXCLIPS_BLOAD, "BL&oad");
  file_menu->Append(WXCLIPS_BSAVE, "BS&ave");
  file_menu->AppendSeparator();
  file_menu->Append(WXCLIPS_QUIT, "&Exit");

  wxMenu *run_menu = new wxMenu;

  run_menu->Append(WXCLIPS_RESETRUN, "R&eset and Run");
  run_menu->Append(WXCLIPS_RUN, "&Run");
  run_menu->Append(WXCLIPS_RUN_LIMIT, "Run with &limit");
  run_menu->Append(WXCLIPS_RESET, "Re&set");
  run_menu->AppendSeparator();
  run_menu->Append(WXCLIPS_CLEAR, "&Clear");
  run_menu->AppendSeparator();
  run_menu->Append(WXCLIPS_WATCHES, "&Watch");
  run_menu->Append(WXCLIPS_ADD_BREAKPOINT, "&Add breakpoint");
  run_menu->Append(WXCLIPS_REMOVE_BREAKPOINT, "Re&move breakpoint");
  run_menu->AppendSeparator();
  run_menu->Append(WXCLIPS_SET_STRATEGY, "&Set strategy");
  run_menu->Append(WXCLIPS_ASSERT_FACT, "Assert &fact");

  wxMenu *app_menu = new wxMenu;
  app_menu->Append(WXCLIPS_CLEAN_WINDOWS, "&Clean up windows");
  app_menu->Append(WXCLIPS_RUN_APP, "&Start Application");

  wxMenu *list_menu = new wxMenu;
  list_menu->Append(WXCLIPS_LIST_RULES, "&Rules");
  list_menu->Append(WXCLIPS_LIST_FACTS, "&Facts");
  list_menu->Append(WXCLIPS_LIST_MATCHES, "&Matches");
  list_menu->Append(WXCLIPS_LIST_AGENDA, "&Agenda");
  list_menu->Append(WXCLIPS_LIST_BREAKPOINTS, "&Breakpoints");
  list_menu->Append(WXCLIPS_LIST_TEMPLATES, "&Templates");
  list_menu->Append(WXCLIPS_LIST_GLOBALS, "&Globals");
  list_menu->Append(WXCLIPS_LIST_STRATEGY, "&Current strategy");
  list_menu->AppendSeparator();
  list_menu->Append(WXCLIPS_CLEAR_SCREEN, "&Clear screen");

  wxMenu *help_menu = new wxMenu;
  help_menu->Append(WXCLIPS_REFERENCE, "&CLIPS for wxWindows Reference");
  help_menu->AppendSeparator();
  help_menu->Append(WXCLIPS_ABOUT, "&About wxCLIPS");

  wxMenuBar *menu_bar = new wxMenuBar;

  menu_bar->Append(file_menu, "&File");
  menu_bar->Append(run_menu, "&Rules");
  menu_bar->Append(app_menu, "&Application");
  menu_bar->Append(list_menu, "&List");
  menu_bar->Append(help_menu, "&Help");

  // Associate the menu bar with the frame
  SetMenuBar(menu_bar);

  // Make a text window
  panel = new wxPanel(this, 0, 0, 500, COMMAND_PANEL_HEIGHT);
  (void)new wxButton(panel, (wxFunction)EnterCommand,
     "Do");
  command_line = new wxMultiText(panel, NULL, "CLIPS> ", "", -1, -1,
                                  COMMAND_TEXT_WIDTH, 90);

  text_window = new wxTextWindow(this, 0, COMMAND_PANEL_HEIGHT, 500, 400, wxBORDER);

  SetSize(x, y, w, h);
  Show(TRUE);
}

wxClipsEnvFrame::~wxClipsEnvFrame(void)
{
  wxTheClipsFrame = NULL;
}

// Intercept menu commands
char *wxBatchFile = NULL;
void wxClipsEnvFrame::OnMenuCommand(int id)
{
  switch (id)
  {
    case WXCLIPS_BATCH:
    {
      char *path = NULL;
      char *name = NULL;
      if (wxBatchFile)
      {
        path = PathOnly(wxBatchFile);
        name = FileNameFromPath(wxBatchFile);
      }

      char *file = wxFileSelector("CLIPS batch file to execute",
                     path, name,
                     "clp", "*.clp", 0, this);
      if (file)
      {
        wxExecuteClipsFile(file);
        if (wxBatchFile) delete wxBatchFile;
        wxBatchFile = copystring(file);
      }
      break;
    }
    case WXCLIPS_LOAD:
    {
      char *file = wxFileSelector("CLIPS file to load",
                     PathOnly(wxBatchFile), NULL,
                     "clp", "*.clp", 0, this);
      if (file)
      {
        SetPrintWhileLoading(CLIPS_TRUE);
        LoadConstructs(file);
        SetPrintWhileLoading(CLIPS_FALSE);
        if (wxBatchFile) delete wxBatchFile;
        wxBatchFile = copystring(file);
      }
      break;
    }
    case WXCLIPS_SAVE:
    {
      char *file = wxFileSelector("CLIPS file to save to",
                     PathOnly(wxBatchFile), NULL,
                     "clp", "*.clp", 0, this);
      if (file)
      {
	SaveConstructs(file);
        if (wxBatchFile) delete wxBatchFile;
        wxBatchFile = copystring(file);
      }
      break;
    }
    case WXCLIPS_BLOAD:
    {
      char *file = wxFileSelector("CLIPS file to binary load",
                     PathOnly(wxBatchFile), NULL,
                     "clp", "*.clp", 0, this);
      if (file)
      {
        SetPrintWhileLoading(CLIPS_TRUE);
	Bload(file);
        SetPrintWhileLoading(CLIPS_FALSE);
        if (wxBatchFile) delete wxBatchFile;
        wxBatchFile = copystring(file);
      }
      break;
    }
    case WXCLIPS_BSAVE:
    {
      char *file = wxFileSelector("CLIPS file to binary save to",
                     NULL, NULL,
                     "clp", "*.clp", 0, this);
      if (file)
      {
	Bsave(file);
        if (wxBatchFile) delete wxBatchFile;
        wxBatchFile = copystring(file);
      }
      break;
    }
    case WXCLIPS_RUN:
    {
      RunCLIPS(-1);
      break;
    }
    case WXCLIPS_RUN_LIMIT:
    {
      char *s = IntToString(limit);
      char *new_s = wxGetTextFromUser("Enter limit value: ",
                            "Respond please",s, this);
      if (new_s)
      {
	StringToInt(new_s, &limit);
	RunCLIPS(limit);
      }
      if (s)
	delete s;
      break;
    }
    case WXCLIPS_RESETRUN:
    {
      ResetCLIPS();
      RunCLIPS(-1);
      break;
    }
    case WXCLIPS_CLEAR:
    {
      ClearCLIPS();
      break;
    }
    case WXCLIPS_RESET:
    {
      ResetCLIPS();
      break;
    }
    case WXCLIPS_WATCHES:
    {
      wxDialogBox dialog(this, "Watches", TRUE, 100, 100, 200, 200);
      (void)new wxMessage(&dialog, "Edit watches");
      dialog.NewLine();

      wxCheckBox check0(&dialog, NULL, "All");
      check0.SetValue(all_watch);

      dialog.NewLine();

      wxCheckBox check1(&dialog, NULL, "Facts");
      check1.SetValue(facts_watch);

      dialog.NewLine();

      wxCheckBox check2(&dialog, NULL, "Rules");
      check2.SetValue(rules_watch);

      dialog.NewLine();

      wxCheckBox check3(&dialog, NULL, "Activations");
      check3.SetValue(activations_watch);

      dialog.NewLine();

      wxCheckBox check4(&dialog, NULL, "Compilations");
      check4.SetValue(compilations_watch);

      dialog.NewLine();

      wxCheckBox check5(&dialog, NULL, "Statistics");
      check5.SetValue(statistics_watch);

      dialog.NewLine();

      wxCheckBox check6(&dialog, NULL, "Globals");
      check6.SetValue(globals_watch);

      dialog.NewLine();
      (void) new wxButton(&dialog, (wxFunction)GenericOk, "Ok");

      dialog.Fit();
      dialog.Show(TRUE);

      all_watch = check0.GetValue();
      SetWatchItem("all", all_watch);

      facts_watch = check1.GetValue();
      SetWatchItem("facts", facts_watch);

      rules_watch = check2.GetValue();
      SetWatchItem("rules", rules_watch);

      activations_watch = check3.GetValue();
      SetWatchItem("activations", activations_watch);

      compilations_watch = check4.GetValue();
      SetWatchItem("compilations", compilations_watch);

      statistics_watch = check5.GetValue();
      SetWatchItem("statistics", statistics_watch);

      globals_watch = check6.GetValue();
      SetWatchItem("globals", globals_watch);

      break;
    }
    case WXCLIPS_ADD_BREAKPOINT:
    {
      char *name = PromptForRuleName();
      if (name)
      {
	void *ptr = FindDefrule(name);
	if (ptr)
	{
	  AddBreakpoint(ptr);
	  char buf[200];
	  sprintf(buf, "Put breakpoint on %s.\n", name);
	  PrintCLIPS("wdialog", buf);
	}
      }
      break;
    }
    case WXCLIPS_REMOVE_BREAKPOINT:
    {
      char *name = PromptForRuleName();
      if (name)
      {
	void *ptr = FindDefrule(name);
	if (ptr)
	{
          RemoveBreakpoint(ptr);
	  char buf[200];
          sprintf(buf, "Removed breakpoint on %s.\n", name);
	  PrintCLIPS("wdialog", buf);
	}
      }
      break;
    }
    case WXCLIPS_ASSERT_FACT:
    {
      char *s = wxGetTextFromUser(
        "Enter fact (no outer brackets): ", "Respond please","", this);
      if (s)
      {
        char buf[300];
        sprintf(buf, "Asserting fact %s\n", s);
        PrintCLIPS("wdialog", buf);
        AssertString(s);
      }
      break;
    }
    case WXCLIPS_LIST_AGENDA:
    {
      ListAgenda();
      break;
    }
    case WXCLIPS_LIST_RULES:
    {
      ListDefrules();
      break;
    }
    case WXCLIPS_LIST_FACTS:
    {
      ListDeffacts();
      break;
    }
    case WXCLIPS_LIST_TEMPLATES:
    {
      ListDeftemplates();
      break;
    }
    case WXCLIPS_LIST_GLOBALS:
    {
      ListDefglobals();
      break;
    }
    case WXCLIPS_LIST_BREAKPOINTS:
    {
      ListBreakpoints();
      break;
    }
    case WXCLIPS_CLEAR_SCREEN:
    {
      text_window->Clear();
      break;
    }
    case WXCLIPS_CLEAN_WINDOWS:
    {
      wxCleanWindows();
      break;
    }
    case WXCLIPS_RUN_APP:
    {
      wxCleanWindows();
      (void)wxStartApp();
      break;
    }
    case WXCLIPS_REFERENCE:
    {
      HelpInstance->LoadFile("clipswin");
      HelpInstance->DisplayContents();
      break;
    }

    case WXCLIPS_ABOUT:
    {
      char aboutBuffer[200];
      sprintf(aboutBuffer, "wxCLIPS\nNASA (c) and Julian Smart, AIAI (c) 1993\nVersion %.2f", wxClipsVersion);
      wxMessageBox(aboutBuffer, "About wxCLIPS");
      break;
    }
    case WXCLIPS_QUIT:
    {
      OnClose();
      delete this;
      break;
    }
  }
}

// Size the subwindows when the frame is resized
void wxClipsEnvFrame::OnSize(int w, int h)
{
  if (panel && text_window)
  {
    int width, height;
    GetClientSize(&width, &height);
    panel->SetSize(0, 0, width, COMMAND_PANEL_HEIGHT);
    text_window->SetSize(0, COMMAND_PANEL_HEIGHT, width, height-COMMAND_PANEL_HEIGHT);
  }
}

// Define the behaviour for the frame closing
// - must delete all frames except for the main one.
Bool wxClipsEnvFrame::OnClose(void)
{
  wxCleanWindows();
  return TRUE;
}

char *wxClipsEnvFrame::PromptForRuleName(void)
{
  wxStringList strings;
  void *ptr = GetNextDefrule(NULL);
  while (ptr)
  {
    char *name = GetDefruleName(ptr);
    strings.Add(name);
    ptr = GetNextDefrule(ptr);
  }
  char **sarray = strings.ListToArray();
  char *the_name = wxGetSingleChoice("Select a rule", "Selection",
       strings.Number(), sarray, this);

  delete sarray;

  return the_name;
}


void GenericOk(wxButton& but, wxCommandEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)but.GetParent();

  dialog->Show(FALSE);
}


// Compatibility
#ifdef wx_msw
int system(char *command)
{
  wxExecute(command);
  return 1;
}
#endif

// I/O Routers
int FindOutput(char *name)
{
  if ((strcmp(name, "wtrace") == 0) ||
      (strcmp(name, "stdout") == 0) ||
      (strcmp(name, "werror") == 0) ||
      (strcmp(name, "wclips") == 0) ||
      (strcmp(name, "wdisplay") == 0) ||
      (strcmp(name, "wagenda") == 0) ||
      (strcmp(name, "wdialog") == 0))
    return (CLIPS_TRUE);
  else
    return (CLIPS_FALSE);
}

void PrintOutput(char *name, char *str)
{
  if (wxTheClipsFrame)
    wxTheClipsFrame->text_window->WriteText(str);
}

void ExitOutput(int exitcode)
{
}

void ClipsYield(void)
{
  (void)wxYield();
}

/*
 * Functions specific to this front end
 *
 */
long clipsQuit()
{
  wxTheClipsFrame->OnClose();
  delete wxTheClipsFrame;
  return 1;
}

/*
 * Required by wx_cmds or CLIPS itself
 *
 */

void ClipsErrorFunction(char *s)
{
  if (wxTheClipsFrame)
    *(wxTheClipsFrame->text_window) << s;
#ifdef wx_x
  else
    cerr << s;
#endif
}

void UserFunctions()
{
  wxUserFunctions();

  DefineFunction("quit", 'l', PTIF clipsQuit, "clipsQuit");
}
